/*
 * $Id: calc.c,v 1.10 2001/04/03 06:22:05 laik Exp $
 *
 * Copyright (C) 1998 Kevin Lai
 *
 * This module contains code for calculating the bandwidth of a 
 * flow given the timings and sizes of its packets.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "kl_util.h"
#include "kl_list.h"

#include "dpcap_flow.h"

#include "calc.h"

/*
 * This structure encapsulates state about the performance calculation.
 */
struct PerfCalc {
    CalcDef *calc_spec;
    PerfCalcSampleSet *sample_set;
    PerfCalcSummaryMetrics summary_metrics;
    int num_sets;
};

const CalcMetricSpec CalcMetricSpecs[] = {
    [CalcMetricSBBW] = {"SBPP", (1 << CalcMetricSBBW)},
    [CalcMetricRBBW] = {"RBPP", (1 << CalcMetricRBBW)},
    [CalcMetricROBW] = {"ROPP", (1 << CalcMetricROBW)},
    [CalcMetricPBW] = {"PBW", (1 << CalcMetricPBW)},
    [CalcMetricABW] = {"ABW", (1 << CalcMetricABW)},
    [CalcMetricArrivalTime] = {"AT", (1 << CalcMetricArrivalTime)},
    [CalcMetricRTT] = {"RTT", (1 << CalcMetricRTT)},
    [CalcMetricDelay] = {"Delay", (1 << CalcMetricDelay)},
    {"", CalcMetricLast},
};

/* XXX Why do we have this? */
inline static int
c_positive_timing(const struct timespec *timing)
{
    return ((KLTimingGetSeconds(timing) > 0) || (KLTimingGetNS(timing) > 0));
}

/* Print packet information. */
static void
c_sample_print(
    KLBufferedOutput *sample_log, const PerfCalcSampleSet *sample,
    long valid_metrics, int bunch_size, const KLListNode *cur_node,
    int transmit_index, int arrival_index, int ack_index)
{
    int i, len;
    DPCapFlowPacket *packet;
    const DPCapFlow *flow;
    KLArray *array;
    char *s, from_time[10], to_time[10];
    DPCapFlowDirection direction;
    struct timespec *timings;

    long from_ns, to_ns;
    struct tm *from_tm, *to_tm;
    time_t time;

    packet = (DPCapFlowPacket *) KLListGetItem(cur_node);
    timings = DPCapFlowPacketTimings(packet);
    flow = DPCapFlowPacketFlow(packet);
    direction = DPCapFlowGetDirection(flow);
    array = KLArrayNew();
    DPCapFlowPacketGetString(packet, array);
    s = KLArrayGetBuffer(array, &len);

    /* Format the time. */
    if (transmit_index != -1) {
	time = KLTimingGetSeconds(&timings[transmit_index]);
	from_tm = gmtime(&time);
	strftime(from_time, sizeof(from_time), "%T", from_tm);
	from_ns = KLTimingGetNS(&timings[transmit_index]) / 1000;
    }
    else {
	from_time[0] = '\0';
	from_ns = 0;
    }
    if (arrival_index != -1) {
	time = KLTimingGetSeconds(&timings[arrival_index]);
	to_tm = gmtime(&time);
	strftime(to_time, sizeof(to_time), "%T", to_tm);
	to_ns = KLTimingGetNS(&timings[arrival_index]) / 1000;
    }
    else {
	to_time[0] = '\0';
	to_ns = 0;
    }
    KLBufferedOutputPrintf(
	sample_log, NULL, "%8s %6ld %8s %6ld %s%3d%3d%3d%2d%2d ",
	from_time, from_ns, to_time, to_ns, s, transmit_index, arrival_index,
	ack_index, direction, bunch_size);
    free(s);
    KLArrayFree(array, FALSE);

    /* Print perf information. */
    for (i = 0; i < CalcMetricLast; i++) {
	if (valid_metrics & CalcMetricSpecs[i].flag) {
	    KLBufferedOutputPrintf(
		sample_log, NULL, "%-4s: %10.5g ",
		CalcMetricSpecs[i].name, sample->d[i]);
	}
    }
    KLBufferedOutputPrintf(sample_log, NULL, "\n");
}

/*
 * This functions decides whether we can calculate a Sender-Based
 * Packet Pair estimate or not.
 */
inline static int
c_sbpp(
    PerfCalcSummaryMetrics *summary_metrics, double transmit_delta,
    double ack_delta)
{
    /* Skip this bunch if there was no RTT */
    if (ack_delta == -1.0) {
	summary_metrics->no_rtt += 1;
	return 1;
    }
	    
    /*
     * Ignore compressed packets and packets for which we have no
     * transmit delta.
     */
    if ((ack_delta < transmit_delta) || (ack_delta == 0.0) ||
	(transmit_delta == -1.0)) {
	summary_metrics->compressed += 1;
	return 1;
    }

    return 0;
}

/*
 * Calculate the round trip time.
 */
inline static long
c_rtt(
    const DPCapFlowPacket *cur_packet, int transmit_index, int ack_index,
    double *rtt)
{
    const struct timespec *cur_timings, *ack_cur_timings;
    struct timespec delta;
    const DPCapFlowPacket *ack_cur_packet;

    /* Calculate RTT only if we have the transmit time and ack time. */
    cur_timings = DPCapFlowPacketTimings(cur_packet);
    ack_cur_packet = DPCapFlowPacketExactAck(cur_packet);
    if ((transmit_index != -1) && (ack_index != -1) &&
	(ack_cur_packet != NULL)) {
	ack_cur_timings = DPCapFlowPacketTimings(ack_cur_packet);
	KLTimingSubtract(&delta, &ack_cur_timings[ack_index], 
		       &cur_timings[transmit_index]);
	*rtt = KLTimingGetDouble(&delta);
	return (1 << CalcMetricRTT);
    }
    return 0;
}

/*
 * Calculate the one way delay.
 */
inline static long
c_delay(
    const DPCapFlowPacket *cur_packet, int transmit_index, int arrival_index,
    double *delay)
{
    const struct timespec *cur_timings;
    struct timespec delta;

    /* Calculate delay only if we have the transmit and arrival time. */
    cur_timings = DPCapFlowPacketTimings(cur_packet);
    if ((transmit_index != -1) && (arrival_index != -1)) {
	KLTimingSubtract(&delta, &cur_timings[arrival_index], 
		       &cur_timings[transmit_index]);
	*delay = KLTimingGetDouble(&delta);
	return (1 << CalcMetricDelay);
    }
    return 0;
}

/*
 * Calculate the arrival time.
 */
inline static long
c_arrival_time(
    const DPCapFlowPacket *cur_packet, int transmit_index, int arrival_index,
    int ack_index, double *arrival_time)
{
    const struct timespec *cur_timings, *ack_cur_timings;
    const DPCapFlowPacket *ack_cur_packet;
    long flags = 0;

    /* Calculate delay only if we have the transmit and arrival time. */
    cur_timings = DPCapFlowPacketTimings(cur_packet);
    ack_cur_packet = DPCapFlowPacketExactAck(cur_packet);
    if (arrival_index != -1) {
	*arrival_time = KLTimingGetDouble(&cur_timings[arrival_index]);
	flags = (1 << CalcMetricArrivalTime);
    }
    else if ((transmit_index != -1) && (ack_index != -1) &&
	(ack_cur_packet != NULL)) {
	ack_cur_timings = DPCapFlowPacketTimings(ack_cur_packet);
	*arrival_time = KLTimingGetDouble(&ack_cur_timings[ack_index]);
	flags = (1 << CalcMetricArrivalTime);
    }
    return flags;
}

/*
 * Calculate delta times. Make sure that deltas are above the
 * clock resolution, otherwise we'd have to divide by 0 later.
 */
inline static void
c_delta_time(
    const CalcDef *cs, const DPCapFlowPacket *first_packet,
    const DPCapFlowPacket *cur_packet, int arrival_index,
    int transmit_index, int ack_index, double *arrival_delta,
    double *transmit_delta, double *ack_delta)
{
    const struct timespec *first_timings, *cur_timings;
    const struct timespec *ack_first_timings, *ack_cur_timings;
    struct timespec delta;
    const DPCapFlowPacket *ack_first_packet, *ack_cur_packet;

    first_timings = DPCapFlowPacketTimings(first_packet);
    cur_timings = DPCapFlowPacketTimings(cur_packet);

    if ((arrival_index != -1) &&
	c_positive_timing(&cur_timings[arrival_index]) &&
	c_positive_timing(&first_timings[arrival_index])) {
	KLTimingSubtract(&delta, &cur_timings[arrival_index],
		       &first_timings[arrival_index]);
	*arrival_delta = KLTimingGetDouble(&delta);
	*arrival_delta = max(*arrival_delta, cs->clock_resolution);
    }
    else {
	*arrival_delta = -1;
    }

    /* Calculate transmit delta times. */
    if ((transmit_index != -1) && 
	c_positive_timing(&cur_timings[transmit_index]) &&
	c_positive_timing(&first_timings[transmit_index])) {
	KLTimingSubtract(&delta, &cur_timings[transmit_index],
		       &first_timings[transmit_index]);
	*transmit_delta = KLTimingGetDouble(&delta);
	*transmit_delta = max(*transmit_delta, cs->clock_resolution);
    }
    else {
	*transmit_delta = -1;
    }

    /*
     * Calculate ack delta times. We must check that the two packets
     * in question were actually acked.
     */
    ack_first_packet = DPCapFlowPacketExactAck(first_packet);
    ack_cur_packet = DPCapFlowPacketExactAck(cur_packet);
    if ((ack_index != -1) && (ack_first_packet != NULL) &&
	(ack_cur_packet != NULL)) {

	ack_first_timings = DPCapFlowPacketTimings(ack_first_packet);
	ack_cur_timings = DPCapFlowPacketTimings(ack_cur_packet);

	/*
	 * This is the difference in the arrival times of two
	 * acknowledgements, not the difference in RTT times.
	 */
	KLTimingSubtract(&delta, &ack_cur_timings[ack_index],
		       &ack_first_timings[ack_index]);
	*ack_delta = KLTimingGetDouble(&delta);
	*ack_delta = max(*ack_delta, cs->clock_resolution);
    }
    else {
	*ack_delta = -1;
    }
}

/*
 * This function calculates a new bandwidth sample based on new packet
 * timing data.
 */
inline static int
c_new_sample(
    const CalcDef *cs, PerfCalcSummaryMetrics *summary_metrics,
    const KLListNode *first_node, const KLListNode *cur_node,
    int transmit_index, int arrival_index,
    int ack_index, PerfCalcSampleSet *sample)
{
    int transmit_size = 0, arrival_size = 0, ack_size = 0, size;
    double arrival_delta, transmit_delta, ack_delta, potential_bw;
    const DPCapFlowPacket *first_packet, *cur_packet, *packet, *ack, *cum_ack;
    const KLListNode *list_node;
    struct timespec *timings, *ack_timings;

    first_packet = (DPCapFlowPacket *) KLListGetItem(first_node);
    cur_packet = (DPCapFlowPacket *) KLListGetItem(cur_node);
    sample->valid_metrics = 0;

    /* Can't do anything else with 1 packet. */
    if (first_node == cur_node) {
	return 1;
    }
    summary_metrics->num_bunches += 1;

    /* First and last packets must be same size */
    if (DPCapFlowPacketGetSize(first_packet) != 
	DPCapFlowPacketGetSize(cur_packet)) {
	summary_metrics->not_same_size += 1;
	return 1; 
    }

    /* Calculate delta times. */
    c_delta_time(
	cs, first_packet, cur_packet, arrival_index, transmit_index,
	ack_index, &arrival_delta, &transmit_delta, &ack_delta);

    /* Add up total bunch size. */
    for (list_node = KLListNext(first_node); list_node != KLListNext(cur_node);
	 list_node = KLListNext(list_node)) {
	packet = (DPCapFlowPacket *) KLListGetItem(list_node);
	timings = DPCapFlowPacketTimings(packet);
	size = DPCapFlowPacketGetSize(packet);
	/* Check that packet was transmitted. */
	if (c_positive_timing(&timings[transmit_index])) {
	    transmit_size += size;
	}
	/* Check that packet actually arrived. */
	if (c_positive_timing(&timings[arrival_index])) {
	    arrival_size += size;
	}
	/* Check that packet was acked. */
	ack = DPCapFlowPacketExactAck(packet);
	cum_ack = DPCapFlowPacketCumAck(packet);
	if (ack != NULL) {
	    ack_timings = DPCapFlowPacketTimings(ack);
	    if (c_positive_timing(&ack_timings[ack_index])) {
		ack_size += size;
	    }
	}
	else if (cum_ack != NULL) {
	    ack_timings = DPCapFlowPacketTimings(cum_ack);
	    if (c_positive_timing(&ack_timings[ack_index])) {
		ack_size += size;
	    }
	}
    }

    /* Receiver-Only Packet Pair bandwidth. */
    if (arrival_delta >= 0.0) {
	sample->d[CalcMetricROBW] = (arrival_size * 8.0) / arrival_delta;
	sample->valid_metrics |= 1 << CalcMetricROBW;
    }
    
    /*
     * Receiver-Based Packet Pair bandwidth.
     * No RBBW for packets that have been compressed. This doesn't
     * get all of them, but it gets the obvious ones.
     */
    if ((arrival_delta - transmit_delta + cs->clock_resolution>= 0.0) &&
	(arrival_delta > 0.0) && (transmit_delta >= 0.0)) {
	sample->d[CalcMetricRBBW] = (arrival_size * 8.0) / arrival_delta;
	sample->valid_metrics |= 1 << CalcMetricRBBW;
    }
    else {
	summary_metrics->compressed += 1;
    }

    /* Sender-Based Packet Pair bandwidth. */
    if (c_sbpp(summary_metrics, transmit_delta, ack_delta) == 0) {
	sample->d[CalcMetricSBBW] = (ack_size * 8.0) / ack_delta;
	sample->valid_metrics |= 1 << CalcMetricSBBW;
    }

    if (transmit_delta >= 0.0) {
	potential_bw = (transmit_size * 8.0) / transmit_delta;
	sample->d[CalcMetricPBW] = potential_bw;
	sample->valid_metrics |= 1 << CalcMetricPBW;
    }
    else {
	sample->d[CalcMetricPBW] = 0.0;
    }
    return 0;
}

/*
 * This function calculates a new performance sample.
 */
int
perf_calc_new_packet(
    PerfCalc *perf_calc, const KLListNode *first_packet_node,
    const KLListNode *cur_packet_node, int transmit_index, int arrival_index,
    int ack_index)
{
    const KLListNode *begin_packet_node;
    long valid_metrics = 0;
    int n, i;
    PerfCalcSampleSet *sample;
    const DPCapFlowPacket *cur_packet;

    if (first_packet_node != cur_packet_node) {
	begin_packet_node = KLListPrev(cur_packet_node);
    }
    else {
	begin_packet_node = first_packet_node;
    }

    /* Calculate non-bunch metrics. */
    cur_packet = (DPCapFlowPacket *) KLListGetItem(cur_packet_node);
    sample = &perf_calc->sample_set[0];
    sample->valid_metrics = c_rtt(
	cur_packet, transmit_index, ack_index, &sample->d[CalcMetricRTT]);
    sample->valid_metrics |= c_delay(
	cur_packet, transmit_index, arrival_index, 
	&sample->d[CalcMetricDelay]);
    sample->valid_metrics |= c_arrival_time(
	cur_packet, transmit_index, arrival_index, ack_index,
	&sample->d[CalcMetricArrivalTime]);

    /* Iterate over different bunch sizes. */
    for (n = perf_calc->calc_spec->min_bunch_size;
	 ((n <= perf_calc->calc_spec->max_bunch_size) &&
	  (begin_packet_node != KLListPrev(first_packet_node))); 
	 n++, begin_packet_node = KLListPrev(begin_packet_node)) {
	
	i = n - perf_calc->calc_spec->min_bunch_size + 1;
	sample = &perf_calc->sample_set[i];

	/* Calculate new sample */
	if (c_new_sample(
	    perf_calc->calc_spec, &perf_calc->summary_metrics,
	    begin_packet_node, cur_packet_node,
	    transmit_index, arrival_index,
	    ack_index, sample) != 1) {
	    valid_metrics |= sample->valid_metrics;
	    if (perf_calc->calc_spec->sample_log != NULL) {
		c_sample_print(
		    perf_calc->calc_spec->sample_log,
		    sample, valid_metrics, n,
		    cur_packet_node, transmit_index, arrival_index, ack_index);
	    }
	}
	/* Flow is too short to handle a larger bunch size. */
	if (begin_packet_node == first_packet_node) {
	    break;
	}
    }
    return valid_metrics;
}

/*
 * This function is called to create a new performance calculation.
 */
PerfCalc *
perf_calc_new(CalcDef *calc_spec)
{
    PerfCalc *perf_calc;

    /* Allocate memory. */
    perf_calc = malloc(sizeof(PerfCalc));

    /* Initialize the summary metrics counts. */
    memset(&perf_calc->summary_metrics, 0,
	   sizeof(struct PerfCalcSummaryMetrics));

    /* Initialize the calc specifications. */
    perf_calc->calc_spec = calc_spec;

    /* The 0th set is for metrics of one packet. */
    perf_calc->num_sets = calc_spec->max_bunch_size - 
	calc_spec->min_bunch_size + 2;
    perf_calc->sample_set = calloc(
	sizeof(PerfCalcSampleSet), perf_calc->num_sets);

    return perf_calc;
}

/*
 * This function is called to delete an old performance calculation.
 */
void
perf_calc_delete(PerfCalc *perf_calc)
{
    free(perf_calc->sample_set);
    free(perf_calc);
}

PerfCalcSampleSet *
perf_calc_get_sample_set(const PerfCalc *perf_calc, int *num_sets)
{
    *num_sets = perf_calc->num_sets;
    return perf_calc->sample_set;
}

const PerfCalcSummaryMetrics *
perf_calc_get_summary_metrics(const PerfCalc *perf_calc)
{
    return &perf_calc->summary_metrics;
}

