/*
 * $Id: filter.c,v 1.20 2001/06/20 23:49:48 laik Exp $
 *
 * This file contains routines which do kernel density estimation.
 *
 * Copyright (C) 1998 Kevin Lai
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */ 

#include <assert.h>
#include <math.h>
#include <string.h>

#include "kl_statistics.h"
#include "kl_statistics_metric_summary.h"
#include "kl_statistics_kernel.h"
#include "kl_util.h"

#include "filter.h"

typedef enum {
    f_kernel_metric_index_measured_bw,
    f_kernel_metric_index_kernel,
    f_kernel_metric_index_measured_bw_mean,
    f_kernel_metric_index_potential_bw_mean,
    f_kernel_metric_index_last,
} f_kernel_metric_index;

typedef struct f_eviction_entry {
    KLTableTuple *tuple; /* Sample tuple */
    int index;
    double utility;      /* Utility of this sample */
} f_eviction_entry;

typedef struct FilterSampleSet {
    Filter *filter;                        /* Parent filter */
    KLStatisticsSet *kernel_source_set;    /* Set of bw, pbw */
    KLStatisticsKernel *kernel;            /* Density kernel */
    KLStatisticsMetricType **metric_types; /* Metric types */
    KLStatisticsMapping *kernel_mapping;   /* Kernel mapping */
    KLStatisticsSet *kernel_set;           /* Statistical distributions */
    unsigned int mapped_dim;               /* Dimensions of kernel set */
    f_eviction_entry *eviction_samples;    /* Eviction samples */
    int correlate;                         /* */
} FilterSampleSet;

/*
 * This structure represents the state kept on all the filtering
 * algorithm.
 */
struct Filter {
    FilterDef def;
    FilterSampleSet **sets;
    double *kernel_resolution;
};

inline static FilterSampleSet *
f_sample_set_new(Filter *filter, int correlate)
{
    KLStatisticsSetDef set_def;
    int *permutation;
    KLStatisticsKernelDef kernel_def;
    FilterSampleSet *f_set;
    unsigned int m;

    f_set = malloc(sizeof(FilterSampleSet));
    f_set->filter = filter;
    f_set->correlate = correlate;
    f_set->mapped_dim = f_kernel_metric_index_last;

    /* Create kernel source set. */
    set_def.dimensions = FilterMetricLast;
    set_def.source_set = NULL;
    set_def.bootstrap = 0;
    f_set->kernel_source_set = KLStatisticsSetNew(&set_def);

    /* Create permutation and metric types */
    permutation = malloc(sizeof(int) * f_set->mapped_dim);
    f_set->metric_types = malloc(
	sizeof(KLStatisticsMetricType *) * f_set->mapped_dim);
    for (m = 0; m < f_set->mapped_dim; m++) {
	permutation[m] = -1;
	f_set->metric_types[m] = NULL;
    }

    /* Only map the measured bandwidth. */
    permutation[f_kernel_metric_index_measured_bw] = FilterMetricMeasuredBW;

    /* Create kernel. */
    kernel_def.source_dim = 1;
    kernel_def.mapped_dim = f_set->mapped_dim;
    kernel_def.permutation = permutation;
    kernel_def.resolution = filter->def.kernel_resolution;
    kernel_def.kernel_function = filter->def.kernel_function;
    kernel_def.kernel_bound_ratio = filter->def.kernel_bound_ratio;
    f_set->kernel = KLStatisticsKernelNew(&kernel_def);

    /* Create mapping. */
    f_set->kernel_mapping = KLStatisticsKernelMappingNew(f_set->kernel);
    
    /* Create kernel metric. */
    f_set->metric_types[f_kernel_metric_index_kernel] = 
	KLStatisticsKernelMetricTypeNew(f_set->kernel);
    f_set->metric_types[f_kernel_metric_index_measured_bw_mean] = 
	KLStatisticsMetricSummaryMeanTypeNew(FilterMetricMeasuredBW);
    f_set->metric_types[f_kernel_metric_index_potential_bw_mean] = 
	KLStatisticsMetricSummaryMeanTypeNew(FilterMetricPotentialBW);

    /* Create kernel set. */
    set_def.dimensions = f_set->mapped_dim;
    set_def.source_set = f_set->kernel_source_set;
    set_def.bootstrap = 0;
    set_def.permutation = permutation;
    set_def.mapping = f_set->kernel_mapping;
    set_def.metric_types = f_set->metric_types;
    f_set->kernel_set = KLStatisticsSetNew(&set_def);

    free(permutation);

    /* Initialize eviction state. */
    f_set->eviction_samples = malloc(
	sizeof(f_eviction_entry) * filter->def.window_size * 2);
    return f_set;
}

inline static void
f_sample_set_delete(FilterSampleSet *f_set)
{
    unsigned int m;

    KLStatisticsSetFree(f_set->kernel_set);
    KLStatisticsMappingFree(f_set->kernel_mapping);
    for (m = 0; m < f_set->mapped_dim; m++) {
	if (f_set->metric_types[m] != NULL) {
	    KLStatisticsMetricTypeFree(f_set->metric_types[m]);
	}
    }
    free(f_set->metric_types);
    KLStatisticsKernelFree(f_set->kernel);
    KLStatisticsSetFree(f_set->kernel_source_set);
    free(f_set->eviction_samples);
    free(f_set);
}

/*
 * Requirements for filtering:
 *
 *   1) Density: More samples -> greater weighted value
 *   2) Potential bandwidth:
 *     a) Potential bw == measured bw -> much smaller weighted value
 *     b) Potential bw == measured bw -> larger measured bw counts for more
 *     c) Same potential bw -> larger measured bw counts for more
 */
inline static double
f_weighted(
    FilterSampleSet *fset, KLTableTuple *tuple, double max_kernel, 
    double log_max_mbw, double log_min_mbw)
{
    const double *tuple_values;
    double weighted_value, kernel, log_mbw, log_pbw;
    double pbw_value, mbw_value, kernel_value;

    tuple_values = KLTableTupleGetValues(tuple);
    kernel = tuple_values[f_kernel_metric_index_kernel];
    kernel_value = tuple_values[f_kernel_metric_index_kernel] / max_kernel;

    if (fset->correlate != -1) {
	log_mbw = max(log(tuple_values[
	    f_kernel_metric_index_measured_bw_mean]), 0);
	mbw_value = (log_mbw - log_min_mbw) / 
	    max((log_max_mbw - log_min_mbw), .000001);
	log_pbw = max(
	    log(tuple_values[f_kernel_metric_index_potential_bw_mean]), 
	    log_mbw);

	pbw_value = 1.0 - (log_mbw / log_pbw);
	weighted_value = 
	    (fset->filter->def.kernel_weight * kernel_value) + 
	    (fset->filter->def.pbw_weight * pbw_value) + 
	    (fset->filter->def.mbw_weight * mbw_value);
#if 0
	printf(
	    "mbw: %f pbw: %f kernel: %f pm_ratio: %f weighted: %f\n",
	    tuple_values[f_kernel_metric_index_measured_bw_mean],
	    tuple_values[f_kernel_metric_index_potential_bw_mean],
	    tuple_values[f_kernel_metric_index_kernel],
	    pm_ratio, weighted_value);
#endif
    }
    else {
	weighted_value = kernel_value;
    }
    return weighted_value;
}

inline static double
f_pmar(
    FilterSampleSet *fset, const double *values, const double *recent_values,
    double log_max_mbw, double log_min_mbw)
{
    double age, log_mbw, log_pbw, pbw_value, mbw_value, age_value;

    log_mbw = max(log(values[FilterMetricMeasuredBW]), 0);
    mbw_value = (log_mbw - log_min_mbw) / 
	max((log_max_mbw - log_min_mbw), .000001);
    log_pbw = max(log(values[FilterMetricPotentialBW]), log_mbw);
    pbw_value = 1.0 - (log_mbw / log_pbw);
    age = recent_values[FilterMetricArrivalTime] - 
	values[FilterMetricArrivalTime];
    age_value = 1 - (log10(age)/log10(1000000));
    return (fset->filter->def.pbw_weight * pbw_value) + 
	(fset->filter->def.mbw_weight * mbw_value) +
	(fset->filter->def.age_weight * age_value);
}

static int
f_eviction_sample_compare(const void *v1, const void *v2)
{
    const f_eviction_entry *e1 = v1, *e2 = v2;
    return (e1->utility < e2->utility) ? -1 : 
	((e1->utility == e2->utility) ? 0 : 1);
}

/*
 * Decide how to evict a sample from the filter.
 */
static void
f_evict(FilterSampleSet *fset, const double *sample)
{    
    unsigned int e, n;
    KLTable *table;
    KLTableTuple *tuple;
    const double *values;
    double log_min_mbw, log_max_mbw;

    table = KLStatisticsSetGetTable(fset->kernel_source_set);
    assert (KLTableCount(table) <= (2 * fset->filter->def.window_size));
    n = KLTableCount(table);
    
    /* Use FIFO eviction when we don't know potential bw. */
    if (fset->correlate == -1) {
	if (n > fset->filter->def.window_size) {
	    tuple = KLTableFirstTuple(table);
	    KLTableDeleteTuple(table, tuple);
	}
	return;
    }

    /* Wait until we hit the high water mark to evict. */
    if (n < (2 * fset->filter->def.window_size)) {
	return;
    }

    /* Get maximum measured bandwidth value. */
    tuple = KLTableSelect(table, FilterMetricMeasuredBW, n - 1);
    log_max_mbw = max(log(KLTableTupleGetValues(tuple)[
 	FilterMetricMeasuredBW]), 0);
 	
    /* Get minimum measured bandwidth value. */
    tuple = KLTableSelect(table, FilterMetricMeasuredBW, 0);
    log_min_mbw = max(log(KLTableTupleGetValues(tuple)[
 	FilterMetricMeasuredBW]), 0);

    /* Recalculate utility */
    for (tuple = KLTableFirstTuple(table), e = 0;
	 tuple != NULL; tuple = KLTableNextTuple(tuple), e++) {
	fset->eviction_samples[e].tuple = tuple;
	values = KLTableTupleGetValues(tuple);
	fset->eviction_samples[e].utility = f_pmar(
	    fset, values, sample, log_max_mbw, log_min_mbw);
    }

    /* Sort samples */
    qsort(
	fset->eviction_samples, n, sizeof(f_eviction_entry), 
	&f_eviction_sample_compare);
    
    /* Evict weakest samples */
    for (e = 0; e < n - fset->filter->def.window_size; e++) {
	tuple = fset->eviction_samples[e].tuple;
#if 0
	values = KLTableTupleGetValues(tuple);
	printf(
	    "evicting mbw: %f pbw: %f age: %f utility: %f\n",
	    values[FilterMetricMeasuredBW],
	    values[FilterMetricPotentialBW],
	    sample[FilterMetricArrivalTime] - values[FilterMetricArrivalTime],
	    fset->eviction_samples[e].utility);
	fflush(stdout);
#endif
	KLTableDeleteTuple(table, tuple);
    }
}

#if 0
/*
 * Decide whether and how to evict a sample from the filter.
 */
static void
f_evict(FilterSampleSet *fset, const double *sample)
{    
    unsigned int e, n, t;
    KLTable *table;
    KLTableTuple *tuple;
    const double *values;
    KLSet *eviction_sample_set;

    if (fset->filter->def.window_size == 0) {
	return;
    }
    table = KLStatisticsSetGetTable(fset->kernel_source_set);
    n = KLTableCount(table);
    if (n < fset->filter->def.window_size) {
	return;
    }

    /* Prevent duplicates */
    eviction_sample_set = KLSetNew(
	f_set->filter->def.window_size, 0, NULL, NULL);

    /* Gather new eviction samples and recalculate utility */
    for (e = 0; e < fset->filter->def.eviction_sample_size; e++) {
	if (!((e <= fset->filter->def.eviction_memory_size) &&
	      (fset->eviction_samples[e].tuple != NULL))) {
	    for (
	    t = 1 + (int) (n * rand() / (RAND_MAX + 1.0));
	    tuple = KLTableSelect(table, 0, t);
	    fset->eviction_samples[e].tuple = tuple;
	    fset->eviction_samples[e].index = t;
	}
	values = KLTableTupleGetValues(fset->eviction_samples[e].tuple);
	fset->eviction_samples[e].utility = f_pmar(values, sample);
    }

    /* Sort samples */
    qsort(
	fset->eviction_samples, fset->filter->def.eviction_sample_size,
	sizeof(f_eviction_entry), &f_eviction_sample_compare);
    
    /* Evict weakest sample */
    tuple = fset->eviction_samples[0].tuple;
    KLTableDeleteTuple(table, tuple);
    fset->eviction_samples[0].tuple = NULL;

    KLSetFree(f_set->eviction_sample_set);
}

#endif

/*
 */
static void 
f_compute_result(FilterSampleSet *fset, FilterResult *result)
{
    f_eviction_entry *sample;
    KLTable *table;
    double total = 0.0, total_square = 0.0, bw, mean, variance;
    unsigned int s, count;
    const double *values;

    /* Fill in result */
    table = KLStatisticsSetGetTable(fset->kernel_set);
    result->num_kernels = KLTableCount(table);

    sample = &fset->eviction_samples[result->num_kernels - 1];
    values = KLTableTupleGetValues(sample->tuple);

    result->bw = values[f_kernel_metric_index_measured_bw_mean];
    result->pbw = values[f_kernel_metric_index_potential_bw_mean];
    result->weight = sample->utility;
    result->kernel = values[f_kernel_metric_index_kernel];

    table = KLStatisticsSetGetTable(fset->kernel_source_set);
    result->num_samples = KLTableCount(table);
    
    // count = min((unsigned int) 5, result->num_kernels);
    count = result->num_kernels;
    for (s = 0; s < count; s++) {
	sample = &fset->eviction_samples[result->num_kernels - s - 1];
	values = KLTableTupleGetValues(sample->tuple);
	bw = values[f_kernel_metric_index_measured_bw_mean];
	total += bw;
	total_square += bw * bw;
    }
    mean = total / s;
    variance = (total_square / s) - (mean * mean);
    result->error = sqrt(max(variance, 0)) / result->bw;
}

/*
 * Initialize filter data structures.
 */
Filter *
filter_new(FilterDef *filter_def)
{
    Filter *filter;
    unsigned int s, size;

    /* Allocate and initialize the filter. */
    filter = malloc(sizeof(Filter));
    memcpy(&filter->def, filter_def, sizeof(FilterDef));
    filter->def.eviction_memory_size = 4;
    filter->def.eviction_sample_size = 12;
    filter->sets = malloc(sizeof(FilterSampleSet *) * filter->def.num_sets);
    for (s = 0; s < filter->def.num_sets; s++) {
	filter->sets[s] = NULL;
    }
    size = sizeof(double) * FilterMetricLast;
    filter->kernel_resolution = malloc(size);
    memcpy(filter->kernel_resolution, filter_def->kernel_resolution, size);

    return filter;
}

/*
 * Free filter data structures.
 */
void
filter_delete(Filter *filter)
{
    unsigned int s;

    for (s = 0; s < filter->def.num_sets; s++) {
	if (filter->sets[s] != NULL) {
	    f_sample_set_delete(filter->sets[s]);
	}
    }
    free(filter->kernel_resolution);
    free(filter->sets);
    free(filter);
}

void
filter_new_metric_set(Filter *filter, int set_index, int correlate)
{
    assert(filter->sets[set_index] == NULL);
    filter->sets[set_index] = f_sample_set_new(filter, correlate);
}

/*
 * This function adds a new sample to the filter set. We make a copy of the
 * PerfSample.
 */
void
filter_add_sample(Filter *filter, int set_index, const double *sample)
{
    FilterSampleSet *fset;
    KLTable *table;

    /* Add tuple to set. */
    fset = filter->sets[set_index];
    table = KLStatisticsSetGetTable(fset->kernel_source_set);
    KLTableNewTuple(table, sample, NULL);

    /* Append to the sample log. */
    if (filter->def.sample_log != NULL) {
	KLArray *array;
	int len;
	array = KLArrayNew();
	KLTableGetString(KLStatisticsSetGetTable(fset->kernel_set), array);
	KLBufferedOutputPrintf(
	    filter->def.sample_log, NULL, "%s",
	    (char *) KLArrayGetBuffer(array, &len));
	KLArrayFree(array, TRUE);
    }
    f_evict(fset, sample);
}

/*
 * This function gets the result of the filtering.
 */
unsigned int 
filter_get_result(
    Filter *filter, int set_index, FilterResult *result)
{
    FilterSampleSet *fset;
    unsigned int num_kernels, k;
    KLTable *table;
    KLTableTuple *tuple;
    double log_min_mbw, log_max_mbw, max_kernel;

    fset = filter->sets[set_index];
    table = KLStatisticsSetGetTable(fset->kernel_set);
    num_kernels = KLTableCount(table);
    if (num_kernels == 0) {
	return 0;
    }

     /* Get maximum kernel value. */
    tuple = KLTableSelect(
    table, f_kernel_metric_index_kernel, num_kernels - 1);
    max_kernel = KLTableTupleGetValues(tuple)[f_kernel_metric_index_kernel];
 	
    /* Get maximum measured bandwidth value. */
    tuple = KLTableSelect(
 	table, f_kernel_metric_index_measured_bw_mean, num_kernels - 1);
    log_max_mbw = max(log(KLTableTupleGetValues(tuple)[
 	f_kernel_metric_index_measured_bw_mean]), 0);
 	
    /* Get minimum measured bandwidth value. */
    tuple = KLTableSelect(
 	table, f_kernel_metric_index_measured_bw_mean, 0);
    log_min_mbw = max(log(KLTableTupleGetValues(tuple)[
 	f_kernel_metric_index_measured_bw_mean]), 0);

    for (tuple = KLTableFirstTuple(table), k = 0;
	 tuple != NULL; tuple = KLTableNextTuple(tuple), k++) {
	fset->eviction_samples[k].tuple = tuple;
	fset->eviction_samples[k].utility = f_weighted(
	    fset, tuple, max_kernel, log_max_mbw, log_min_mbw);
    }

    /* Sort samples */
    qsort(
	fset->eviction_samples, num_kernels, sizeof(f_eviction_entry), 
	&f_eviction_sample_compare);

    f_compute_result(fset, result);
    return 1;
}

/*
 * This function verifies the structure of the filter.
 */
void
filter_verify(Filter *filter)
{
    assert(filter != NULL);
}
