/*
 * $Id: flow_calc.c,v 1.32 2001/04/16 22:54:35 laik Exp $
 *
 * This modules takes packet information from the dpcap_flow module
 * and passes it on to the calc module.
 *
 * Copyright (C) 1998 Kevin Lai
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
#include <assert.h>
#include <ctype.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "kl_util.h"

#include "flow_calc.h"
#include "filter.h"

typedef struct FlowCalcMetric {
    int pp_metric_index;       /* Indices into sample of metrics. */
    int correlating_index;     /* Indices of correlating metrics. */
    int valid;
} FlowCalcMetric;

/*
 * This encapsulates the state of one flow calculation.
 */
struct FlowCalcFlow {
    FlowCalcSet *set;           /* Parent set. */
    Filter *pp_filter;          /* Filter for the packet pair metrics. */
    FlowCalcMetric *metrics;   
    int pp_row_size;            /* Number of packet pair filter rows. */
    int pp_num_metrics;         /* Number of packet pair_metrics. */
    const DPCapFlow *dpcap_flow;/* Flow we belong to. */
    PerfCalc *perf_calc;        /* Performance calculation state. */
    KLListNode *list_node;      /* KLList node in the flow calc set. */
    char src_name[NI_MAXHOST];  /* Name of the source. */
    char dst_name[NI_MAXHOST];  /* Name of the destination. */
    const struct in_addr *flow_src;
    const struct in_addr *flow_dst;
    int timer_dst;              /* Timer index of dst */
    int timer_src;              /* Timer index of src */
};

/*
 * This encapsulates the flow calculation state. It keeps track
 * of all the calculation state for all of the the flows.
 */
struct FlowCalcSet {
    FlowCalcSetDef def;
    int gethost_failed;              /* Whether gethostbyaddr has failed. */
    KLList *flow_calc_list;          /* KLList of all the flow_calcs. */
};

/*
 * The idea here is to orthogonalize how we calculate the
 * different bandwidths algorithms. We always want to calculate RBBW
 * over ROBW when we have a choice. In many cases, we don't have a
 * choice because of the set of timings we have. For example, RBBW needs
 * timings from one timer to another. SBBW needs timings from one timer
 * back to itself. ROBW only needs timings from one timer.
 *
 * The idea is to form a table indexed by the timer's indices and each
 * element dictates which algorithm should be used and whether potential
 * bandwidth filtering should be used. This is for 3 timers:
 *
 *              to    The row index is the index of the src timer + 1
 *           0  1  2  because the src timer could be -1 (for no src 
 * f -1 (0) ro ro ro  timer). The column index is just the dest timer
 * r  0 (1) sb rb rb  index.
 * o  1 (2) rb sb rb
 * m  2 (3) rb rb sb
 *
 * The table is actually kept as two C arrays: pp_metric_index and
 * correlating_index.
 */
static void
fc_init_packet_pair_filter(FlowCalcFlow *flow, int num_timers)
{
    int m;
    CalcDef *cs;
    double kernel_resolution[2];
    FilterDef fdef;
    FlowCalcMetric *metric;

    /*
     * Calculate number of metrics. There is one less column than
     * row because ROPP is encoded at -1 -> n, but there is no n -> -1.
     */
    cs = &flow->set->def.calc_spec;
    flow->pp_row_size = num_timers;
    flow->pp_num_metrics = num_timers * (num_timers + 1);
    kernel_resolution[0] = cs->kernel_resolution;
    kernel_resolution[1] = cs->kernel_resolution;

    /* Initialize the filter. */
    fdef.num_sets = flow->pp_num_metrics;
    fdef.window_size = cs->window_size;
    fdef.kernel_bound_ratio = cs->kernel_bound_ratio;
    fdef.kernel_resolution = kernel_resolution;
    fdef.kernel_function = cs->kernel_function;
    fdef.kernel_weight = cs->kernel_weight;
    fdef.pbw_weight = cs->pbw_weight;
    fdef.mbw_weight = cs->mbw_weight;
    fdef.age_weight = cs->kernel_weight;
    fdef.sample_log = flow->set->def.sample_log;

    flow->pp_filter = filter_new(&fdef);

    /* Initialize metric indices. */
    flow->metrics = malloc(sizeof(FlowCalcMetric) * flow->pp_num_metrics);

    for (m = 0; m < flow->pp_num_metrics; m++) {
	metric = &flow->metrics[m];
	metric->valid = 1;
	metric->correlating_index = CalcMetricPBW;

	/* Handle ROPP cases. */
	if (m < flow->pp_row_size) {
	    metric->pp_metric_index = CalcMetricROBW;
	    metric->correlating_index = -1;
	    if (flow->timer_src == m) {
		metric->valid = 0;
	    }
	}
	/* Handle SBPP cases. */
	else if ((m % (flow->pp_row_size + 1)) == flow->pp_row_size) {
	    metric->pp_metric_index = CalcMetricSBBW;
	    if (flow->timer_dst == (m % flow->pp_row_size)) {
		metric->valid = 0;
	    }
	}
	else {
	    metric->pp_metric_index = CalcMetricRBBW;
	}
	filter_new_metric_set(flow->pp_filter, m, metric->correlating_index);
    }
}

/*
 * Get host names from addresses.
 */
inline static void
fc_get_host_names(
    FlowCalcSet *flow_calc_set, 
    const struct in_addr *flow_src, const struct in_addr *flow_dst,
    char *flow_src_buf, int flow_src_buf_len, 
    char *flow_dst_buf, int flow_dst_buf_len)
{
    int flags = 0;
    struct sockaddr_in src_sockaddr, dst_sockaddr;

    flags = flow_calc_set->def.no_convert_addresses == 1 ? NI_NUMERICHOST : 
	flags;
    flags = flow_calc_set->def.qualify_host_names == 0 ? NI_NOFQDN : flags;

    /* XXX glibc 2.2 getnameinfo() is so broken. */
    if (flow_calc_set->def.no_convert_addresses == 1) {
	inet_ntop(AF_INET, flow_src, flow_src_buf, flow_src_buf_len);
	inet_ntop(AF_INET, flow_dst, flow_dst_buf, flow_dst_buf_len);
	return;
    }

    src_sockaddr.sin_family = AF_INET;
    memcpy(&src_sockaddr.sin_addr, flow_src, sizeof(struct in_addr));
    getnameinfo(
	(struct sockaddr *) &src_sockaddr, sizeof(src_sockaddr), 
	flow_src_buf, flow_src_buf_len, NULL, 0, flags);

    dst_sockaddr.sin_family = AF_INET;
    memcpy(&dst_sockaddr.sin_addr, flow_dst, sizeof(struct in_addr));
    getnameinfo(
	(struct sockaddr *) &dst_sockaddr, sizeof(dst_sockaddr), 
	flow_dst_buf, flow_dst_buf_len, NULL, 0, flags);

    /* XXX glibc 2.2 getnameinfo is so broken. */
    if (flow_calc_set->def.qualify_host_names == 0) {
	char *s;
	s = strchr(flow_src_buf, '.');
	if ((s != NULL) && (!isdigit(flow_src_buf[0]))) {
	    *s = '\0';
	}
	s = strchr(flow_dst_buf, '.');
	if ((s != NULL) && (!isdigit(flow_dst_buf[0]))) {
	    *s = '\0';
	}
    }
}

/*
 * Initialize the flow calc flow.
 */
static FlowCalcFlow *
fc_flow_new(FlowCalcSet *set, DPCapFlow *dpcap_flow)
{
    FlowCalcFlow *flow;
    const DPCapFlowSet *flow_set;
    int num_timers;

    flow = malloc(sizeof(FlowCalcFlow));
    flow->set = set;

    /* Initialize perf calc state. */
    flow->perf_calc = perf_calc_new(&set->def.calc_spec);
    flow->dpcap_flow = dpcap_flow;
    
    flow_set = DPCapFlowGetSet(dpcap_flow);
    num_timers = DPCapFlowSetNumTimers(flow_set);

    /* Set flow endpoints' names. */
    DPCapFlowGetEndPoints(
	dpcap_flow, &flow->flow_src, &flow->timer_src, &flow->flow_dst, 
	&flow->timer_dst);
    fc_get_host_names(
	set, flow->flow_src, flow->flow_dst, flow->src_name, 
	sizeof(flow->src_name), flow->dst_name, sizeof(flow->dst_name));

    /* Initialize filtering. */
    fc_init_packet_pair_filter(flow, num_timers);

    return flow;
}

static void
fc_flow_delete(void *client_data, FlowCalcFlow *flow)
{
    assert(client_data == NULL);
    free(flow->metrics);
    filter_delete(flow->pp_filter);
    perf_calc_delete(flow->perf_calc);
    free(flow);
}

/*
 * Verify the structure of the flow.
 */
static void
fc_flow_verify(FlowCalcFlow *flow)
{
    filter_verify(flow->pp_filter);
}

inline static KLException *
fc_append_log(
    FlowCalcFlow *flow, struct timespec *from_timings, int from_index,
    struct timespec *to_timings, int to_index, const DPCapFlowPacket *packet)
{
    char from_time[10], to_time[10];
    const char *metric_name;
    int packet_id;
    long from_ns, to_ns;
    FilterResult result;
    struct tm *from_tm, *to_tm;
    time_t time;
    const DPCapFlow *dpcap_flow;
    KLException *ex = NULL;
    const KLNetworkProtocolPacket *np_p;
    const struct ip *ip;

    /* Get flow src, dst, direction. */
    dpcap_flow = flow->dpcap_flow;

    np_p = DPCapFlowPacketGetPacket(packet);
    ip = (const struct ip *) (np_p->buffer + np_p->header_offsets[0]);
    packet_id = ntohs(ip->ip_id);

    /* Format the time. */
    if (from_index != -1) {
	time = KLTimingGetSeconds(&from_timings[from_index]);
	from_tm = gmtime(&time);
	strftime(from_time, sizeof(from_time), "%T", from_tm);
	from_ns = KLTimingGetNS(&from_timings[from_index]) / 1000;
    }
    else {
	from_time[0] = '0';
	from_time[1] = '\0';
	from_ns = 0;
    }
    time = KLTimingGetSeconds(&to_timings[to_index]);
    to_tm = gmtime(&time);
    strftime(to_time, sizeof(to_time), "%T", to_tm);
    to_ns = KLTimingGetNS(&to_timings[to_index]) / 1000;

    if (flow_calc_get_metrics(
	flow, from_index, to_index, &metric_name, &result) == 0) {
	return NULL;
    }

    if (flow->set->def.calc_spec.verbose == 0) {
	ex = KLBufferedOutputPrintf(
	    flow->set->def.result_log, NULL, 
	    "%8s.%06ld %8s.%06ld %-15s %-15s %2d %d %5d %s %13.0f %13.0f %8.3g\n",
	    from_time, from_ns, to_time, to_ns, flow->src_name, flow->dst_name,
	    from_index, to_index, 
	    packet_id, metric_name, result.bw, result.pbw, result.error);
    }
    else {
	ex = KLBufferedOutputPrintf(
	    flow->set->def.result_log, NULL, 
	    "%8s.%06ld %8s.%06ld %-15s %-15s %2d %d %5d %s %13.0f %13.0f %8.3g %5.3f %7.4g %5d %9d\n",
	    from_time, from_ns, to_time, to_ns, flow->src_name, flow->dst_name,
	    from_index, to_index, 
	    packet_id, metric_name, result.bw, result.pbw, result.error, 
	    result.weight, result.kernel, result.num_kernels, 
	    result.num_samples);
    }
    return ex;
}

/*
 * Add some metrics to the the flow_calc's filters. Determine if we
 * have a valid sample based on the 
 */
inline static KLException *
fc_add_sample(
    FlowCalcFlow *flow, const DPCapFlowPacket *packet,
    struct timespec *from_timings, int from_index, struct timespec *to_timings,
    int to_index)
{
    int s, num_sets, set_index;
    int valid_pp_sample = 0;
    PerfCalcSampleSet *sample_set_array, *set;
    double sample[FilterMetricLast];
    KLException *ex = NULL;
    FlowCalcMetric *metric;

    set_index = to_index + (flow->pp_row_size * (from_index + 1));
    metric = &flow->metrics[set_index];
    if (metric->valid == 0) {
	return ex;
    }

    /* Add sample to the packet pair filter. */
    sample_set_array = perf_calc_get_sample_set(flow->perf_calc, &num_sets);

    /*
     * Iterate over different bunch sizes and take the first valid sample.
     */
    for (s = 0; s < num_sets; s++) {
	set = &sample_set_array[s];
	valid_pp_sample = set->valid_metrics & 
	    CalcMetricSpecs[metric->pp_metric_index].flag &
	    flow->set->def.calc_spec.restrict_packet_pair;
	if (valid_pp_sample) {
	    sample[FilterMetricMeasuredBW] = set->d[metric->pp_metric_index];
	    if (metric->correlating_index != -1) {
		sample[FilterMetricPotentialBW] = 
		    set->d[metric->correlating_index];
	    }
	    else {
		sample[FilterMetricPotentialBW] = 0.0;
	    }
	    if (sample_set_array[0].valid_metrics & 
		(1 << CalcMetricArrivalTime)) {
		sample[FilterMetricArrivalTime] = 
		    sample_set_array[0].d[CalcMetricArrivalTime];
	    }
	    else {
		sample[FilterMetricArrivalTime] = 0.0;
	    }
	    filter_add_sample(flow->pp_filter, set_index, sample);
	    /* Use the first valid sample. */
	    break;
	}
    }

    if (valid_pp_sample && (flow->set->def.result_log != NULL)) {
	ex = fc_append_log(
	    flow, from_timings, from_index, to_timings, to_index,
	    packet);
    }

    return ex;
}

/*
 * Calculate the performance using the ack. We use the new_index
 * as the transmit_index because it only makes sense to calculate
 * performance from a timer back to the same timer when using
 * ack timings.
 */
inline static KLException *
fc_ack(
    FlowCalcFlow *flow_calc, const KLListNode *first_node,
    const KLListNode *cur_node, const DPCapFlowPacket *packet,
    int timer_index)
{
    const DPCapFlowPacket *ack_packet;
    struct timespec *ack_timings, *timings;
    long valid_metrics;
    KLException *ex = NULL;

    ack_packet = DPCapFlowPacketExactAck(packet);
    ack_timings = DPCapFlowPacketTimings(ack_packet);
    timings = DPCapFlowPacketTimings(packet);

    valid_metrics = perf_calc_new_packet(
	flow_calc->perf_calc, first_node, cur_node, timer_index, -1,
	timer_index);
    ex = KLExceptionPass(NULL, fc_add_sample(
	flow_calc, packet, timings, timer_index, ack_timings,
	timer_index));
    return ex;
}

/*
 * Handle a new packet.
 */
inline static KLException *
fc_packet_new(
    FlowCalcFlow *flow_calc, const KLListNode *first_node,
    const KLListNode *cur_node, const DPCapFlowPacket *packet,
    int timer_index)
{
    struct timespec *timings;
    int prev_index, next_index;
    long valid_metrics;
    KLException *ex = NULL;

    timings = DPCapFlowPacketTimings(packet);

    /* Compute the index of the closest previous and next timings. */
    DPCapFlowPacketFindTimings(packet, timer_index, &prev_index, &next_index);

    /* Calculate performance from the previous timer to this one. */
    valid_metrics = perf_calc_new_packet(
	flow_calc->perf_calc, first_node, cur_node, prev_index,
	timer_index, -1);
    ex = fc_add_sample(
	flow_calc, packet, timings, prev_index, timings, timer_index);
    if (ex != NULL) {
	return ex;
    }
    /* Calculate performance from this timer to the next one. */
    if (next_index != -1) {
	valid_metrics = perf_calc_new_packet(
	    flow_calc->perf_calc, first_node, cur_node, timer_index,
	    next_index, -1);
	ex = fc_add_sample(
	    flow_calc, packet, timings, timer_index, timings, next_index);
	if (ex != NULL) {
	    return ex;
	}
    }
    return NULL;
}

/*
 * Write out the stats on one flow.
 */
inline static KLException *
fc_update_file_flow(
    KLBufferedOutput *bo, FlowCalcFlow *flow, int from_index, int to_index)
{
    const PerfCalcSummaryMetrics *sm;
    const char *metric_name;
    FilterResult result;
    KLException *ex = NULL;

    if (flow_calc_get_metrics(
	flow, from_index, to_index, &metric_name, &result) == 0) {
	return NULL;
    }
    sm = perf_calc_get_summary_metrics(flow->perf_calc);
    if (flow->set->def.calc_spec.verbose == 0) {
	ex = KLExceptionPass(NULL, KLBufferedOutputPrintf(
	    bo, NULL, "%-15.15s %-15.15s %2d %2d %4s %13.0f %13.0f %8.3g\n",
	    flow->src_name, flow->dst_name, from_index, to_index,
	    metric_name, result.bw, result.pbw, result.error));
    } else {
	ex = KLExceptionPass(NULL, KLBufferedOutputPrintf(
	    bo, NULL,
	"%-15.15s %-15.15s %2d %2d %4s %13.0f %13.0f %8.3g %9d %9d %9d %9d %5.3f %7.2g %5d %9d\n",
	flow->src_name, flow->dst_name, from_index, to_index,
	metric_name, result.bw, result.pbw, result.error, sm->not_same_size,
	sm->compressed,
	sm->no_rtt, sm->num_bunches, result.weight, result.kernel, 
	result.num_kernels, result.num_samples));
    }
    return ex;
}

/*
 * Create and initialize a flow_calc data structure.
 */
FlowCalcSet *
flow_calc_set_new(FlowCalcSetDef *def)
{
    FlowCalcSet *set;

    set = malloc(sizeof(FlowCalcSet));
    memcpy(&set->def, def, sizeof(FlowCalcSetDef));
    set->def.calc_spec.sample_log = set->def.sample_log;
    set->gethost_failed = 0;

    /* Initialize the flow calc flow list. */
    set->flow_calc_list = KLListNew();

    return set;
}

/*
 * Delete a flow_calc data structure.
 */
void
flow_calc_set_delete(FlowCalcSet *set)
{
    /* Delete flow calc flow list. */
    KLListFree(
	set->flow_calc_list, (KLListFreeFun) fc_flow_delete, NULL);

    free(set);
}

/*
 * Handle new packet information. Separate flows into forward and
 * backward flows (relative to the ordering of the timers).
 */
KLException *
flow_calc_set_packet_new(
    DPCapFlow *dpcap_flow, const KLListNode *cur_node, int timer_index,
    int ack, void *flow_set_client_data)
{
    FlowCalcSet *flow_calc_set;
    FlowCalcFlow *flow;
    const DPCapFlowPacket *packet;
    KLListNode *first_node;
    const DPCapFlowSet *flow_set;
    KLException *ex = NULL;

    /* Extract all relevant data from the parameters. */
    flow_calc_set = flow_set_client_data;
    packet = (DPCapFlowPacket *) KLListGetItem(cur_node); 
    first_node = KLListFirst(DPCapFlowGetPacketList(dpcap_flow));
    flow_set = DPCapFlowGetSet(dpcap_flow);
    flow = DPCapFlowGetClientData(dpcap_flow);

#if 0
    {
	const KLNetworkProtocolPacket *p;
	p = DPCapFlowPacketGetPacket(packet);
	if (((const struct iphdr *)p->buffer)->id == htons(52276)) {
	    printf("hi\n");
	}
    }
#endif

    /* We keep one set of data for each direction. */
    if (ack) {
	ex = fc_ack(
	    flow, first_node, cur_node, packet, timer_index);
    }
    else {
	ex = fc_packet_new(
	    flow, first_node, cur_node, packet, timer_index);
    }

    /* Keep the list in LIFO order. */
    KLListFreeNode(flow_calc_set->flow_calc_list, flow->list_node);
    flow->list_node = KLListPrependItem(
	flow_calc_set->flow_calc_list, flow);

    if (flow_calc_set->def.packet_new != NULL) {
	ex = flow_calc_set->def.packet_new(
	    flow, cur_node, timer_index, ack, flow_calc_set->def.client_data);
    }
    return ex;
}

/*
 * Create and initialize a new flow.
 */
void
flow_calc_set_flow_new(DPCapFlow *dpcap_flow, void *flow_set_client_data)
{
    FlowCalcSet *flow_calc_set;
    FlowCalcFlow *flow_calc;

    assert(dpcap_flow != NULL);
    flow_calc_set = (FlowCalcSet *) flow_set_client_data;

    /* Create new flow calc flow. */
    flow_calc = fc_flow_new(flow_calc_set, dpcap_flow);

    /* Add it to the list. */
    flow_calc->list_node = KLListPrependItem(
	flow_calc_set->flow_calc_list, flow_calc);

    /* Set state in dpcap. */
    DPCapFlowSetClientData(dpcap_flow, flow_calc);
}

/*
 * Write out the state of all the flows.
 */
KLException *
flow_calc_update_output(KLBufferedOutput *bo, void *client_data)
{
    FlowCalcSet *flow_calc_set;
    KLListNode *list_node;
    FlowCalcFlow *flow;
    int from_index, to_index, pp_row_size;
    KLException *ex = NULL;

    flow_calc_set = client_data;
    ex = KLExceptionPass(NULL, KLBufferedOutputPrintf(
	bo, NULL, "Cmds: (q)uit, any other key to update\n"));
    if (flow_calc_set->def.calc_spec.verbose == 0) {
	ex = KLExceptionPass(NULL, KLBufferedOutputPrintf(
	    bo, NULL,
	    "%-15s %-15s %2s %2s %4s %13s %13s %8s\n",
	    "FlowSource", "FlowDest", "FI", "TI", "Metr", "Bandwidth", 
	    "Sent BW",
	    "Error"));
    }
    else {
	ex = KLExceptionPass(NULL, KLBufferedOutputPrintf(
	    bo, NULL,
      "%-15s %-15s %2s %2s %4s %13s %8s %9s %9s %9s %9s %5s %7s %5s %9s\n",
	    "FlowSource", "FlowDest", "FI", "TI", "Metr", "Bandwidth", "Error",
	    "WrongSize", "Comp", "NoRTT", "Bunches", "W",
	    "Kernel", "NKern", "NSamples"));
    }
    if (ex != NULL) {
	return ex;
    }
    KLListLoop(list_node, flow_calc_set->flow_calc_list) {
	flow = (FlowCalcFlow *) KLListGetItem(list_node);
	pp_row_size = flow->pp_row_size;
	for (from_index = -1; from_index < pp_row_size; from_index++) {
	    for (to_index = 0; to_index < pp_row_size; to_index++) {
		ex = fc_update_file_flow(bo, flow, from_index, to_index);
		if (ex != NULL) {
		    return ex;
		}
	    }
	}
    }
    return NULL;
}

/*
 * Delete a flow.
 */
void
flow_calc_set_flow_delete(DPCapFlow *dpcap_flow, void *flow_set_client_data)
{
    FlowCalcSet *flow_calc_set;
    FlowCalcFlow *flow;

    flow_calc_set = (FlowCalcSet *) flow_set_client_data;
    flow = DPCapFlowGetClientData(dpcap_flow);

    /* Delete from flow list. */
    KLListFreeNode(flow_calc_set->flow_calc_list, flow->list_node);

    /* Delete flow. */
    fc_flow_delete(NULL, flow);
}

/*
 * Verify the flow_calc_set data structure.
 */
void
flow_calc_set_verify(void *client_data)
{
    FlowCalcSet *set;
    KLListNode *list_node;
    FlowCalcFlow *flow;

    set = (FlowCalcSet *) client_data;

    KLListLoop(list_node, set->flow_calc_list) {
	flow = (FlowCalcFlow *) KLListGetItem(list_node);
	fc_flow_verify(flow);
    }
}

#if 0
FlowCalcFlow *
flow_calc_set_get_flow(
    FlowCalcSet *flow_calc_set, const struct in_addr *flow_src,
    const struct in_addr *flow_dst, FlowCalcDeleteFlowFunc delete_flow,
    void *client_data)
{
    FlowCalcFlow *flow;
    KLListNode *list_node;
    int found = 0;

    KLListLoop(list_node, flow_calc_set->flow_calc_list) {
	/* Get flow src, dst, direction. */
	flow = (FlowCalcFlow *) KLListGetItem(list_node);
	if ((flow_src->s_addr == flow->flow_src->s_addr) &&
	    (flow_dst->s_addr == flow->flow_dst->s_addr)) {
	    found = 1;
	    break;
	}
    }

    if (!found) {
	return NULL;
    }
    flow->delete_flow = delete_flow;
    flow->client_data = client_data;
    return flow;
}
#endif

/*
 * Get the metrics for a particular flow. 0 indicated no result.
 * 1 indicates a result.
 */
int
flow_calc_get_metrics(
    FlowCalcFlow *flow, int from_index, int to_index,
    const char **metric_name, FilterResult *result)
{
    unsigned int set_index, rval = 0;
    FlowCalcMetric *metric;

    set_index = to_index + (flow->pp_row_size * (from_index + 1));
    metric = &flow->metrics[set_index];
    if (metric->valid == 1) {
	*metric_name = CalcMetricSpecs[metric->pp_metric_index].name;
	rval = filter_get_result(flow->pp_filter, set_index, result);
    }
    return rval;
}


