/*
 * $Id: lbc_packet_size_range.c,v 1.3 2001/07/19 21:52:20 laik Exp $
 *
 * Copyright (C) 1998 Kevin Lai
 *
 * This module contains code for calculating the sigma bandwidth.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>

#include "kl_array.h"
#include "kl_statistics.h"
#include "kl_statistics_metric_summary.h"
#include "kl_statistics_kernel.h"
#include "kl_statistics_mapping_1to1.h"
#include "kl_statistics_regression.h"
#include "kl_socket.h"

#include "transport.h"
#include "lbc_sigma.h"

enum {
    lbcs_source_metric_sent_size = 0,
    lbcs_source_metric_received_size,
    lbcs_source_metric_delay,
    lbcs_source_metric_last,
};

enum {
    lbcs_calc_metric_sent_size = 0,
    lbcs_calc_metric_received_size,
    lbcs_calc_metric_min_delay,
    lbcs_calc_metric_kernel,
    lbcs_calc_metric_last,
};

/*
 * Maintains information about link bandwidth calculation probe packets.
 */
typedef struct lbc_packet_size_range_t {
    KLStatisticsSet *source_set;         /* Contains packet delay and size */
    KLStatisticsKernel *kernel;          /* Used to do min filtering */
    KLStatisticsMapping *mapping;        /* Used to do min filtering */
    KLStatisticsMetricType *metric_types[lbcs_calc_metric_last];
                                         /* Used to do min filtering */
    KLStatisticsSet *min_filtered_set;   /* Contains min filtered data */
    KLStatisticsSet *regression_set;     /* Contains regression data */
    KLStatisticsMapping *regression_mapping; 
                                         /* Regression mapping. */
    KLStatisticsRegression *sr;          /* Used for regression. */

    unsigned int num_packet_sizes;       /* Number of different packet sizes */
    unsigned int receive_per_size;
    unsigned int min_packet_size;        /* Min packet size on path */
    unsigned int max_packet_size;        /* Max packet size on path */
    double spacing;
    unsigned int *packets_received;
} lbc_packet_size_range_t;

static void
lpsr_resize(
    lbc_packet_size_range_t *ps_range, unsigned int num_packet_sizes, 
    unsigned int receive_per_size, unsigned int min_packet_size,
    unsigned int max_packet_size)
{

    ps_range->num_packet_sizes = num_packet_sizes;
    ps_range->receive_per_size = receive_per_size;
    ps_range->min_packet_size = min_packet_size;
    ps_range->max_packet_size = max_packet_size;
    ps_range->spacing = 
	(ps_range->max_packet_size - ps_range->min_packet_size) /
	(ps_range->num_packet_sizes - 1.0);
    if (ps_range->packets_received != NULL) {
	free(ps_range->packets_received);
    }
    ps_range->packets_received = calloc(
	num_packet_sizes, sizeof(unsigned int));
}

void
lbc_packet_size_range_resize(
    lbc_packet_size_range_t *ps_range, unsigned int min_packet_size,
    unsigned int max_packet_size)
{
    lpsr_resize(
	ps_range, ps_range->num_packet_sizes, ps_range->receive_per_size, 
	min_packet_size, max_packet_size);
}

lbc_packet_size_range_t *
lbc_packet_size_range_new(
    unsigned int num_packet_sizes, unsigned int receive_per_size,
    unsigned int min_packet_size, unsigned int max_packet_size)
{
    lbc_packet_size_range_t *ps_range;
    int permutation[lbcs_calc_metric_last] = {0, 1, -1, -1};
    int regression_permutation[2] = 
    {lbcs_calc_metric_received_size, lbcs_calc_metric_min_delay};
    double resolution[2] = {1, 1};
    KLStatisticsSetDef set_def;
    KLStatisticsKernelDef kernel_def;
    KLStatisticsRegressionDef sr_def;
    KLStatisticsMetricType *metric_types[2] = {NULL, NULL};
    KLStatisticsMapping1to1Def mapping_1to1_def;

    ps_range = malloc(sizeof(lbc_packet_size_range_t));

    /* Create source set. */
    /* Dimensions: sent size, received size, delay. */
    set_def.dimensions = lbcs_source_metric_last;
    set_def.source_set = NULL;
    set_def.bootstrap = 0;
    ps_range->source_set = KLStatisticsSetNew(&set_def);

    /* Create min filtered set. */
    kernel_def.source_dim = lbcs_source_metric_last;
    kernel_def.mapped_dim = lbcs_calc_metric_last;
    kernel_def.permutation = permutation;
    kernel_def.resolution = resolution;
    kernel_def.kernel_function = KLStatisticsKernelFunctionUniform;
    kernel_def.kernel_bound_ratio = 0.0;
    ps_range->kernel = KLStatisticsKernelNew(&kernel_def);

    /* Create metric types. */
    ps_range->metric_types[lbcs_calc_metric_sent_size] = NULL;
    ps_range->metric_types[lbcs_calc_metric_received_size] = NULL;
    ps_range->metric_types[lbcs_calc_metric_min_delay] = 
	KLStatisticsMetricSummaryMinTypeNew(lbcs_source_metric_delay);
    ps_range->metric_types[lbcs_calc_metric_kernel] = 
	KLStatisticsKernelMetricTypeNew(ps_range->kernel);
    
    /* Create mapping. */
    ps_range->mapping = KLStatisticsKernelMappingNew(ps_range->kernel);

    set_def.dimensions = lbcs_calc_metric_last;
    set_def.permutation = permutation;
    set_def.mapping = ps_range->mapping;
    set_def.metric_types = ps_range->metric_types;
    set_def.source_set = ps_range->source_set;
    set_def.bootstrap = 0;
    ps_range->min_filtered_set = KLStatisticsSetNew(&set_def);

    mapping_1to1_def.permutation = regression_permutation;
    mapping_1to1_def.mapped_dim = 2;
    mapping_1to1_def.source_dim = 4;
    ps_range->regression_mapping = 
	KLStatisticsMapping1to1New(&mapping_1to1_def);

    set_def.dimensions = 2;
    set_def.permutation = regression_permutation;
    set_def.mapping = ps_range->regression_mapping;
    set_def.metric_types = metric_types;
    set_def.source_set = ps_range->min_filtered_set;
    set_def.bootstrap = 0;
    ps_range->regression_set = KLStatisticsSetNew(&set_def);

    /* Delay as a function of size. */
    sr_def.model_set = ps_range->regression_set;
    sr_def.dep_var_index = 1;
    ps_range->sr = KLStatisticsRegressionNew(&sr_def);

    ps_range->packets_received = NULL;
    lpsr_resize(
	ps_range, num_packet_sizes, receive_per_size, min_packet_size,
	max_packet_size);
    return ps_range;
}

void
lbc_packet_size_range_free(lbc_packet_size_range_t *ps_range)
{
    unsigned int m;

    free(ps_range->packets_received);
    KLStatisticsRegressionFree(ps_range->sr);
    KLStatisticsMappingFree(ps_range->regression_mapping);
    KLStatisticsSetFree(ps_range->regression_set);
    KLStatisticsSetFree(ps_range->min_filtered_set);

    for (m = 0; m < lbcs_calc_metric_last; m++) {
	if (ps_range->metric_types[m] != NULL) {
	    KLStatisticsMetricTypeFree(ps_range->metric_types[m]);
	}
    }
    KLStatisticsMappingFree(ps_range->mapping);
    KLStatisticsKernelFree(ps_range->kernel);
    KLStatisticsSetFree(ps_range->source_set);
    free(ps_range);
}


static void
lpsr_received(lbc_packet_size_range_t *ps_range, int size)
{
    unsigned int index;

    index = rint((size - ps_range->min_packet_size) / ps_range->spacing);
    assert(index < ps_range->num_packet_sizes);
    ps_range->packets_received[index] = ps_range->packets_received[index] + 1;
    /* printf("received index: %d\n", index); */
}

void
lbc_packet_size_range_add_sample(
    lbc_packet_size_range_t *ps_range, const DPCapFlowPacket *packet,
    const struct timespec *delay)
{
    double sample[lbcs_source_metric_last];
    const DPCapFlowPacket *ack;
    int ack_size = 0, size;

    size = DPCapFlowPacketGetSize(packet);
    lpsr_received(ps_range, size);
    ack = DPCapFlowPacketExactAck(packet);
    ack_size = DPCapFlowPacketGetSize(ack);

    /* We take into account the size of the ack. */
    sample[lbcs_source_metric_sent_size] = size * 8;
    sample[lbcs_source_metric_received_size] = (size + ack_size) * 8;
    sample[lbcs_source_metric_delay] = KLTimingGetDouble(delay);

#if 0
    printf(
	"sent size: %d received size: %d value %f\n", 
	(int) sample[0], (int) sample[1], sample[2]);
#endif
    KLTableNewTuple(KLStatisticsSetGetTable(ps_range->source_set),
		  sample, NULL);
#if 0
    printf("Source set:\n");
    lbcs_print_set(ps_range->source_set);
#endif
#if 0
    printf("Min Filtered set:\n");
    lbcs_print_set(ps_range->min_filtered_set);
#endif
}

/*
 * Get latest result from the min regression filter.
 */
int
lbc_packet_size_range_regression(
    lbc_packet_size_range_t *ps_range, double *slope, double *intercept,
    double *confidence)
{
    unsigned int num_sizes;
    double coefficients[2], standard_error[2];
    KLStatisticsRegressionSampleStatistics srss;

    /* Need at least 3 samples. */
    num_sizes = KLTableCount(KLStatisticsSetGetTable(
	ps_range->min_filtered_set));
    if (num_sizes < 3) {
	return 1;
    }
    /* KLStatisticsSetRecompute(ps_range->regression_set); */
#if 0
    printf("Regression set:\n");
    lbcs_print_set(ps_range->regression_set);
#endif
    KLStatisticsRegressionCompute(
	ps_range->sr, coefficients, standard_error, &srss);
    *intercept = coefficients[0];
    *slope = coefficients[1];
    *confidence = srss.adj_r_square;
    return 0;
}

void
lbc_packet_size_range_find_min_count(
    lbc_packet_size_range_t *ps_range, unsigned int *packet_count,
    unsigned int *packet_size)
{
    unsigned int min_count, min_packet_count, s, min_index;

    min_packet_count = ULONG_MAX;
    min_count = 0;
    min_index = 0;
    for (s = 0; s < ps_range->num_packet_sizes; s++) {
	if (ps_range->packets_received[s] == min_packet_count) {
	   if ((rand() / (RAND_MAX + 1.0)) < 0.50) {
	       min_index = s;
	       min_packet_count = ps_range->packets_received[s];
	   }
	    min_count++;
	}
	if (ps_range->packets_received[s] < min_packet_count) {
	    min_index = s;
	    min_packet_count = ps_range->packets_received[s];
	    min_count = 0;
	}
    }
    /* Check if we've exhausted the current set of sizes. */
    if (min_packet_count > ps_range->receive_per_size) {
	lpsr_resize(
	    ps_range, ps_range->num_packet_sizes * 2 - 1,
	    ps_range->receive_per_size * 16,
	    ps_range->min_packet_size, ps_range->max_packet_size);
    }
    /* printf("sent index: %d sent count: %d\n", min_index, min_packet_count); */
    *packet_count = min_packet_count;
    *packet_size = rint(min_index * ps_range->spacing) + 
	ps_range->min_packet_size; 
}

/*
 * Search the data for sizes which are off the estimated line.
 */
void
lbc_packet_size_range_find_off_line(
    lbc_packet_size_range_t *ps_range, double slope, double intercept, 
    int *off_line_size)
{
    int num_sizes, s, size, sent_size, received_size;
    double line_delay, delay, max_delay_diff;
    const double *values;
    KLTable *table;
    KLTableTuple *tuple;

    table = KLStatisticsSetGetTable(ps_range->min_filtered_set);
    num_sizes = KLTableCount(table);

    /* Iterate over the different packet sizes. */
    max_delay_diff = -HUGE_VAL;
    size = 0;
    for (s = 0; s < num_sizes; s++) {
	
	/* Get the smallest sample for each value. */
	tuple = KLTableSelect(table, 0, s);
	values = KLTableTupleGetValues(tuple);
	sent_size = values[lbcs_calc_metric_sent_size] / 8;
	received_size = values[lbcs_calc_metric_received_size] / 8;
	delay = values[lbcs_calc_metric_min_delay];
	line_delay = (slope * received_size) + intercept;
	if (delay - line_delay > max_delay_diff) {
	    size = sent_size;
	    max_delay_diff = delay - line_delay;
	}
    }
    *off_line_size = size;
}

/*
 * Get the stats as a string.
 */
void
lbc_packet_size_range_get_string(
    lbc_packet_size_range_t *ps_range, KLArray *array)
{
    KLTable *table;

    table = KLStatisticsSetGetTable(ps_range->min_filtered_set);
    KLTableGetString(table, array);
}
