/*
 * $Id: lbc_sigma.c,v 1.8 2001/07/19 21:52:20 laik Exp $
 *
 * Copyright (C) 1998 Kevin Lai
 *
 * This module contains code for calculating the sigma bandwidth.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>

#include "kl_array.h"
#include "kl_socket.h"

#include "transport.h"
#include "lbc_sigma.h"
#include "lbc_packet_size_range.h"

/*
 * State associated with a sigma calculation session between a source
 * and destination.
 */
struct LBCSigma {
    LBCSigmaDef def;                    /* Definitions. */
    KLEventDefRegistration *timer_reg;    /* For the timer. */

    TransportSession *transport_session;/* Used to send/receive packets. */
    TransportBuffer *buf;

    unsigned int packets_sent;          /* # packets sent this phase. */
    unsigned int packets_received;      /* # packets received this phase. */
    unsigned short packet_id;           /* Next packet id to send. */
    lbc_packet_size_range_t *ps_range;  /* Contains packet size range data */

    double slope;
    double intercept;
    double confidence;
};

/*
 * Allocate and initialize link bandwidth calculation sigma state.
 */
LBCSigma *
lbc_sigma_new(LBCSigmaDef *sigma_def)
{
    LBCSigma *lbc_sigma;
    TransportSessionDef ts_def;
    unsigned int min_packet_size, max_packet_size;

    /* Initialize parameters. */
    lbc_sigma = malloc(sizeof(LBCSigma));
    memcpy(&lbc_sigma->def, sigma_def, sizeof(LBCSigmaDef));

    /* Create new transport session. */
    memcpy(&ts_def.source, &lbc_sigma->def.flow.source,
	   sizeof(TransportAddress));
    memcpy(&ts_def.dest, &lbc_sigma->def.flow.dest,
	   sizeof(TransportAddress));
    KLExceptionErrorCatch(transport_new_session(
	&ts_def, &lbc_sigma->transport_session), "");
    lbc_sigma->buf = transport_new_buffer(lbc_sigma->transport_session);

    transport_session_get_packet_sizes(
	lbc_sigma->transport_session, IPPROTO_TCP, &min_packet_size,
	&max_packet_size);

    /* Initialize packet counters. */
    lbc_sigma->packets_sent = 0;
    lbc_sigma->packets_received = 0;
    lbc_sigma->packet_id = 0;
    lbc_sigma->ps_range = lbc_packet_size_range_new(
	lbc_sigma->def.num_packet_sizes, lbc_sigma->def.receive_per_size,
	min_packet_size, max_packet_size);

    return lbc_sigma;
}

/*
 * Deallocate the sigma resources.
 */
void
lbc_sigma_free(LBCSigma *lbc_sigma)
{
#if 0
    KLArray *array;
    int len;

    array = KLArrayNew();
    lbc_packet_size_range_get_string(lbc_sigma->ps_range, array);
    printf("%s\n", (char *) KLArrayGetBuffer(array, &len));
    KLArrayFree(array, TRUE);
#endif
    lbc_packet_size_range_free(lbc_sigma->ps_range);
    transport_free_buffer(lbc_sigma->buf);
    transport_delete_session(lbc_sigma->transport_session);
    free(lbc_sigma);
}

/*
 * Get the sigma statistics.
 */
void
lbc_sigma_get_statistics(
    LBCSigma *lbc_sigma, LBCSigmaStatistics *lbc_sigma_statistics,
    KLArray *array)
{
    lbc_sigma_statistics->slope = lbc_sigma->slope;
    lbc_sigma_statistics->intercept = lbc_sigma->intercept;
    lbc_sigma_statistics->confidence = lbc_sigma->confidence;
    if (array != NULL) {
	lbc_packet_size_range_get_string(lbc_sigma->ps_range, array);
    }
}

/*
 * Send a sigma probe packet.
 */
int
lbc_sigma_send_packet(LBCSigma *lbc_sigma)
{
    unsigned int packet_size, old_packet_size, packet_count;
    unsigned int min_packet_size, max_packet_size;

    /*
     * Probablistically try to correct any anomalous samples.
     * Don't do this all the time because some anomalous samples are very
     * difficult to correct or can't be corrected.
     */
    packet_size = 0;
    if ((rand() / (RAND_MAX + 1.0)) < 0.25) {
	lbc_packet_size_range_find_off_line(
	    lbc_sigma->ps_range, lbc_sigma->slope, lbc_sigma->intercept, 
	    &packet_size);
    }
    if (packet_size == 0) {
	/* Try to find size with smallest count. */
	lbc_packet_size_range_find_min_count(
	    lbc_sigma->ps_range, &packet_count, &packet_size);
    }

    /* Send packet. */
    old_packet_size = packet_size;
    transport_compose_packet(
	lbc_sigma->transport_session, &lbc_sigma->def.flow.dest.sock_addr,
	IPPROTO_TCP, 255, packet_size, lbc_sigma->packet_id, lbc_sigma->buf);
    packet_size = transport_send_packet(
	lbc_sigma->transport_session, lbc_sigma->buf);
#if 0
    printf("Sent sigma packet #%d size %d id %d.\n",
	   lbc_sigma->packets_sent, packet_size, lbc_sigma->packet_id);
#endif
    lbc_sigma->packet_id++;
    lbc_sigma->packets_sent++;

    /* Check for path MTU change. */
    if (packet_size != old_packet_size) {
	transport_session_get_packet_sizes(
	    lbc_sigma->transport_session, IPPROTO_TCP, &min_packet_size,
	    &max_packet_size);
	lbc_packet_size_range_resize(
	    lbc_sigma->ps_range, min_packet_size, packet_size);
	    
    }
    return 1;
}

/*
 * Handle the reception of a new sample.
 */
int
lbc_sigma_sample_received(
    LBCSigma *lbc_sigma, const DPCapFlowPacket *packet, 
    const struct timespec *delay, unsigned int timing_type)
{
    unsigned short packet_id;
    int rval;
    const u_int8_t *payload;
    const KLNetworkProtocolPacket *np_p;

    /* Only care about full timings. */
    if (timing_type == 0) {
	return 0;
    }

    /* Check if this sample is one of ours. */
    np_p = DPCapFlowPacketGetPacket(packet);
    payload = transport_check_packet(lbc_sigma->transport_session, np_p);
    if (payload == NULL) {
	return 0;
    }
    
    lbc_sigma->packets_received++;
    packet_id = transport_get_packet_id(
	lbc_sigma->transport_session, &lbc_sigma->def.flow.dest.sock_addr,
	np_p);

#if 0
    printf("packet id: %d\n", packet_id);
#endif

    /* Add sample to packet info. */
    lbc_packet_size_range_add_sample(lbc_sigma->ps_range, packet, delay);

    /* Check to see if we have the required confidence. */
    rval = lbc_packet_size_range_regression(
	lbc_sigma->ps_range, &lbc_sigma->slope,
	&lbc_sigma->intercept, &lbc_sigma->confidence);

#if 0
    /* Check for valid estimate. */
    if (rval != 0) {
	printf("no estimate.\n");
    }
#endif

    /*
     * If we've received the last packet we sent, send another one now.
     */
    if (packet_id == (lbc_sigma->packet_id - 1)) {
	/* printf("Got last packet.\n"); */
	return 1;
    }
    return 0;
}
