/*
 * $Id: lbc_tailgate.c,v 1.7 2001/07/19 21:52:20 laik Exp $
 *
 * Copyright (C) 1998 Kevin Lai
 *
 * This module contains code for calculating the sigma bandwidth.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <arpa/inet.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>

#include "kl_array.h"
#include "kl_statistics.h"
#include "kl_statistics_metric_summary.h"
#include "kl_statistics_mapping_summary.h"

#include "transport.h"
#include "lbc_packet_size_range.h"
#include "lbc_tailgate_packet.h"
#include "lbc_tailgate.h"

typedef struct lbct_link {
    struct in_addr addr;            /* Address of far endpoint of link */
    int addr_set;                   /* Whether the address has been set */
    int ttl;                        /* Time-to-live of link */
    unsigned int packets_received;  /* Packets received for link */
    struct timespec min_delay;      /* The minimum delay of tailgaters */
    KLStatisticsSet *source_set;    /* Contains delay samples. */
    lbc_packet_size_range_t *ps_range;

    double min_delay_std_dev;       /* */
    double min_delay_mean;          /* */
    double min_delay_error;         /* */  
    double inv_bw;                  /* 1 / bandwidth */
    int ack_size;                   /* The size of acks in bits */
    unsigned short md0;
    unsigned short md1;
    unsigned short md_ack;
} lbct_link;

struct LBCTailgate {
    LBCTailgateDef def;           /* Definitions. */

    TransportSession *transport_session;/* Used to send/receive packets. */
    TransportBuffer *buf0;
    TransportBuffer *buf1;

    int max_packet_size;
    int min_packet_size;
    unsigned int packets_sent;          /* # packets sent this phase. */
    unsigned int packets_received;      /* # packets received this phase. */
    unsigned int packet_timeouts;       
    unsigned short packet_id;           /* Next packet id to send. */

    KLList *link_list;
    int dest_found;
    KLList *packet_group_list;
};

/*
 * Search the tailgate packet list for a packet id.
 */
static KLListNode *
lbct_packet_group_search(LBCTailgate *lbc_tailgate, unsigned short packet_id)
{
    KLListNode *list_node;
    LBCTailgatePacketGroup *pg;

    for (list_node = KLListLast(lbc_tailgate->packet_group_list); 
	 !KLListFinished(lbc_tailgate->packet_group_list, list_node); 
	 list_node = KLListPrev(list_node)) {
	pg = KLListGetItem(list_node);
	if (lbc_tailgate_packet_group_match(pg, packet_id)) {
	    return list_node;
	}
    }

    return NULL;
}

/*
 * Initialize packet sending state.
 */
static lbct_link *
lbct_link_new(
    int ttl, unsigned int min_packet_size, unsigned int max_packet_size)
{
    lbct_link *link;
    KLStatisticsSetDef set_def;
    unsigned int packet_size_limit;

    link = malloc(sizeof(lbct_link));
    link->ttl = ttl;
    KLTimingSet(&link->min_delay, LONG_MAX, 0);
    link->min_delay_std_dev = 0.0;
    link->min_delay_error = 0.0;
    link->packets_received = 0;
    link->inv_bw = 0.0;
    link->ack_size = 0;
    link->addr_set = 0;

    /* Allocate source set. */
    set_def.dimensions = 1;
    set_def.source_set = NULL;
    set_def.bootstrap = 0;
    link->source_set = KLStatisticsSetNew(&set_def);

    packet_size_limit = min_packet_size * 8;
    link->ps_range = lbc_packet_size_range_new(
	50, 1024, min_packet_size, packet_size_limit);

    return link;
}

static void
lbct_link_free(void *client_data, void *item)
{
    lbct_link *link = item;
    assert(client_data == NULL);
    KLStatisticsSetFree(link->source_set);
    lbc_packet_size_range_free(link->ps_range);
    free(link);
}

static void
lbct_packet_group_free(void *client_data, void *item)
{
    LBCTailgatePacketGroup *pg = item;
    
    assert(client_data == NULL);
    lbc_tailgate_packet_group_free(pg);
}

/*
 * Search the links for the link with the specified ttl.
 */
static lbct_link *
lbct_link_search(KLList *link_list, int ttl)
{
    int i = 1;
    lbct_link *link = NULL;
    KLListNode *list_node;

    KLListLoop(list_node, link_list) {
	if (i == ttl) {
	    link = (lbct_link *) KLListGetItem(list_node);
	    break;
	}
	i++;
    }
    return link;
}

/* Bootstrap to compute error. */
static void
lbct_link_bootstrap(
    LBCTailgate *lbc_tailgate, lbct_link *link)
{
    KLTableTuple *tuple;
    KLTable *source_table, *delay_table;
    KLStatisticsMapping *delay_stats_mapping;
    KLStatisticsMetricType *delay_stats_smt[2];
    KLStatisticsSet *delay_set, *delay_stats_set;
    int permutation[3] = {-1, -1, -1}, num_min_delays, d;
    KLStatisticsMappingSummaryDef mapping_def;
    KLStatisticsSetDef set_def;
    const double *values;

    /* Don't bootstrap unless we have enough samples. */
    source_table = KLStatisticsSetGetTable(link->source_set);
    if (KLTableCount(source_table) < 4) {
	return;
    }

    /* Set up delay set. */
    set_def.dimensions = 1;
    set_def.source_set = NULL;
    set_def.bootstrap = 0;
    delay_set = KLStatisticsSetNew(&set_def);
    delay_table = KLStatisticsSetGetTable(delay_set);
    
    /* Allocate min delay set. */
    mapping_def.source_dim = 1;
    mapping_def.mapped_dim = 2;
    delay_stats_mapping = KLStatisticsMappingSummaryNew(&mapping_def);

    delay_stats_smt[0] = KLStatisticsMetricSummaryMeanTypeNew(0);
    delay_stats_smt[1] = KLStatisticsMetricSummaryVarianceTypeNew(0);

    set_def.dimensions = 2;
    set_def.source_set = delay_set;
    set_def.bootstrap = 0;
    set_def.permutation = permutation;
    set_def.mapping = delay_stats_mapping;
    set_def.metric_types = delay_stats_smt;
    delay_stats_set = KLStatisticsSetNew(&set_def);

    num_min_delays = min(max(0.10 * KLTableCount(source_table), 10U), 
			 min(KLTableCount(source_table), 100U));
    /* printf("ttl: %d num_min_delays:%d\n", link->ttl, num_min_delays); */
    for (d = 0; d < num_min_delays; d++) {
	tuple = KLTableSelect(source_table, 0, d);
	KLTableNewTuple(delay_table, KLTableTupleGetValues(tuple), NULL);
    }

    /* Get delay statistics. */
    tuple = KLTableFirstTuple(KLStatisticsSetGetTable(delay_stats_set));
    values = KLTableTupleGetValues(tuple);
    link->min_delay_mean = values[0];
    link->min_delay_std_dev = sqrt(values[1]);
    link->min_delay_error = link->min_delay_std_dev / link->min_delay_mean;

    /* Free things. */
    KLStatisticsSetFree(delay_stats_set);
    KLStatisticsMetricTypeFree(delay_stats_smt[0]);
    KLStatisticsMetricTypeFree(delay_stats_smt[1]);
    KLStatisticsMappingFree(delay_stats_mapping);
    KLStatisticsSetFree(delay_set);
}

static void
lbct_link_calc(
    LBCTailgate *lbc_tailgate, lbct_link *link, double sigma_slope,
    double sigma_intercept)
{
    double partial_sigma, queueing_delay, min_delay;
    KLListNode *list_node;
    lbct_link *cur_link = NULL;
    int size0, size1, size2;

    partial_sigma = 0.0;

    /* XXX maybe should add from smallest to largest? */
    KLListLoop(list_node, lbc_tailgate->link_list) {
	cur_link = KLListGetItem(list_node);
	if (cur_link == link) {
	    break;
	}
	if (cur_link->inv_bw > 0) {
	    partial_sigma += cur_link->inv_bw;
	}
    }
    assert(!KLListFinished(lbc_tailgate->link_list, list_node));
    size0 = lbc_tailgate->max_packet_size * 8;
    size1 = lbc_tailgate->min_packet_size * 8;
    size2 = cur_link->ack_size * 8;
    min_delay = KLTimingGetDouble(&cur_link->min_delay);
    queueing_delay = min_delay + ((size1 - size2) * sigma_slope) +
	((size1 - size0) * partial_sigma) - 
	(size1 * sigma_slope) - sigma_intercept;
    cur_link->inv_bw = queueing_delay / size0;
}

/*
 * Account for a sample for a link.
 */
inline static void
lbct_link_add_sample(lbct_link *link, LBCTailgatePacketGroup *pg)
{
    struct timespec delay;
    double delay_d;
    const DPCapFlowPacket *last_packet;

    lbc_tailgate_packet_group_get_delay(pg, &delay);
    last_packet = lbc_tailgate_packet_group_get_packet(pg);
    delay_d = KLTimingGetDouble(&delay);
    KLTableNewTuple(
	KLStatisticsSetGetTable(link->source_set), &delay_d, NULL);

    lbc_packet_size_range_add_sample(
	link->ps_range, last_packet, &delay);

    if (KLTimingCompare(&delay, &link->min_delay) < 0) {
	KLTimingCopy(&link->min_delay, &delay);
#if 0
	link->md0 = DPCapFlowPacketId(first->dpcap_packet);
	link->md1 = DPCapFlowPacketId(last->dpcap_packet);
	ack = DPCapFlowPacketExactAck(last->dpcap_packet);
	link->md_ack = DPCapFlowPacketId(ack);
#endif
    }
}

/*
 * Account for reception of tailgated packet's ack.
 */
inline static void
lbct_link_set_addr(
    LBCTailgate *lbc_tailgate, lbct_link *link, LBCTailgatePacketGroup *pg)
{
    lbct_link *cur_link;
    KLListNode *list_node;
    char old_addr[INET_ADDRSTRLEN], new_addr[INET_ADDRSTRLEN];
    const struct in_addr *addr, *dest;

    addr = lbc_tailgate_packet_group_get_addr(pg);
    if (addr == NULL) {
	return;
    }
    if (link->addr_set == 0) {
	/* Check if we've reached the dest. */
	dest = &(((struct sockaddr_in *)
		 &lbc_tailgate->def.flow.dest.sock_addr)->sin_addr);
	if (memcmp(addr, dest, sizeof(struct in_addr)) == 0) {
	    lbc_tailgate->dest_found = 1;
	    list_node = KLListLast(lbc_tailgate->link_list);
	    cur_link = KLListGetItem(list_node);
	    while (cur_link != link) {
		list_node = KLListPrev(list_node);
		KLListFreeNode(lbc_tailgate->link_list, KLListNext(list_node));
		lbct_link_free(NULL, cur_link);
		cur_link = KLListGetItem(list_node);
	    }
	}
    }
    /* Verify that the address hasn't changed. */
    else {
	if (memcmp(&link->addr, addr, sizeof(struct in_addr)) != 0) {
	    inet_ntop(AF_INET, &link->addr, old_addr, sizeof(old_addr));
	    inet_ntop(AF_INET, addr, new_addr, sizeof(new_addr));
	    printf("The address of node %d changed from %16s to %16s.\n",
		   link->ttl, old_addr, new_addr);
	}
    }
    memcpy(&link->addr, addr, sizeof(struct in_addr));
    link->addr_set = 1;
}

/*
 * Check if we're interested in this packet. 
 */
static KLListNode *
lbct_check_packet(
    LBCTailgate *lbc_tailgate, const DPCapFlowPacket *packet,
    unsigned short packet_id, unsigned int timing_type)
{
    const u_int8_t *payload;
    KLListNode *list_node;
    const KLNetworkProtocolPacket *np_p;

    /* Check if this sample is one of ours. */
    np_p = DPCapFlowPacketGetPacket(packet);
    payload = transport_check_packet(lbc_tailgate->transport_session, np_p);
    if (payload == NULL) {
	return NULL;
    }
    
    list_node = lbct_packet_group_search(lbc_tailgate, packet_id);
    /* Check if this is an old packet. */
    if (list_node == NULL) {
	printf("Could not find packet group.\n");
	return NULL;
    }

    return list_node;
}

/*
 * Select a link to probe.
 */
inline static unsigned int
lbct_select_link(LBCTailgate *lbc_tailgate)
{
    KLListNode *list_node;
    lbct_link *link, *prev_link, *bad_link, *min_link, *max_error_link;
    int ttl = 0, num_links;

    num_links = KLListCount(lbc_tailgate->link_list);
    /*
     * Probablistically try to correct link with anomalous min delay.
     * Don't do this all the time because some anomalous samples are very
     * difficult to correct or can't be corrected.
     */
    if ((rand() / (RAND_MAX + 1.0)) < 0.75) {
	max_error_link = KLListGetItem(KLListFirst(lbc_tailgate->link_list));
	prev_link = NULL;
	bad_link = NULL;
	KLListLoop(list_node, lbc_tailgate->link_list) {
	    link = KLListGetItem(list_node);
	    if (prev_link != NULL) {
		if (KLTimingCompare(&prev_link->min_delay,
				  &link->min_delay) >= 0) {
		    bad_link = prev_link;
		    break;
		}
	    }
	    if (link->min_delay_error > max_error_link->min_delay_error) {
		max_error_link = link;
	    }
	    prev_link = link;
	}
	ttl = (bad_link != NULL) ? bad_link->ttl : 
	    (max_error_link != NULL) ? max_error_link->ttl : 0;
    }
    if (ttl == 0) {
	if (!lbc_tailgate->dest_found) {
	    ttl = num_links + 1;
	}
	else {
	    min_link = KLListGetItem(KLListFirst(lbc_tailgate->link_list));
	    KLListLoop(list_node, lbc_tailgate->link_list) {
		link = KLListGetItem(list_node);
		if (link->packets_received < min_link->packets_received) {
		    min_link = link;
		}
	    }
	    ttl = min_link->ttl;
	}
#if 0
	/* Randomly choose a ttl from 1 to 1 + current path length. */
	/* printf("Choosing random link:"); */
	ttl = (int) ((num_links + 1 - lbc_tailgate->dest_found) * 
		     (rand() / (RAND_MAX + 1.0)) + 1);
#endif
    }
    /* printf(" %d\n", ttl); */
    return ttl;
}

static void
lbct_link_get_endpoints(
    LBCTailgate *lbc_tailgate, KLListNode *link_list_node, 
    char *source, unsigned int source_len,
    char *dest, unsigned int dest_len)
{
    lbct_link *link, *prev_link;
    KLListNode *prev_node;

    link = KLListGetItem(link_list_node);

    /* Get names of link endpoints */
    source[0] = '\0';
    dest[0] = '\0';
    if (link_list_node != KLListFirst(lbc_tailgate->link_list)) {
	prev_node = KLListPrev(link_list_node);
	prev_link = KLListGetItem(prev_node);
	if (prev_link->addr_set) {
	    inet_ntop(AF_INET, &prev_link->addr, source, source_len);
	}
    }
    else {
	inet_ntop(AF_INET,
		  &((struct sockaddr_in *)
		    &lbc_tailgate->def.flow.source.sock_addr)->sin_addr,
	    source, source_len);
    }
    if (link->addr_set) {
	inet_ntop(AF_INET, &link->addr, dest, dest_len);
    }
}

/*
 * Allocate and initialize a new link bandwidth calculation tailgate state.
 */
LBCTailgate *
lbc_tailgate_new(LBCTailgateDef *def)
{
    LBCTailgate *lbc_tailgate;
    TransportSessionDef ts_def;

    /* Initialize parameters. */
    lbc_tailgate = malloc(sizeof(LBCTailgate));
    memcpy(&lbc_tailgate->def, def, sizeof(LBCTailgateDef));

    /* Create new transport session. */
    memcpy(&ts_def.source, &lbc_tailgate->def.flow.source,
	   sizeof(TransportAddress));
    memcpy(&ts_def.dest, &lbc_tailgate->def.flow.dest,
	   sizeof(TransportAddress));
    KLExceptionErrorCatch(
	transport_new_session(&ts_def, &lbc_tailgate->transport_session),
	"");
    lbc_tailgate->buf0 = transport_new_buffer(lbc_tailgate->transport_session);
    lbc_tailgate->buf1 = transport_new_buffer(lbc_tailgate->transport_session);

    transport_session_get_packet_sizes(
	lbc_tailgate->transport_session, IPPROTO_TCP,
	&lbc_tailgate->min_packet_size, &lbc_tailgate->max_packet_size);

    /* Initialize packet counters. */
    lbc_tailgate->packets_sent = 0;
    lbc_tailgate->packets_received = 0;
    lbc_tailgate->packet_timeouts = 0;
    lbc_tailgate->packet_id = 0;

    /* Initialize tailgate state. */
    lbc_tailgate->link_list = KLListNew();
    lbc_tailgate->dest_found = 0;
    lbc_tailgate->packet_group_list = KLListNew();

    return lbc_tailgate;
}

/*
 * Free resources associated with lbc tailgate.
 */
void
lbc_tailgate_free(LBCTailgate *lbc_tailgate)
{
    KLListFree(lbc_tailgate->packet_group_list, lbct_packet_group_free, NULL);
    KLListFree(lbc_tailgate->link_list, lbct_link_free, NULL);
    transport_free_buffer(lbc_tailgate->buf0);
    transport_free_buffer(lbc_tailgate->buf1);
    transport_delete_session(lbc_tailgate->transport_session);
    free(lbc_tailgate);
}

/*
 * Send tailgate packets.
 */
unsigned int
lbc_tailgate_send_packet(LBCTailgate *lbc_tailgate)
{
    int size, tailgated_ttl, i;
    unsigned int tailgater_size, packet_count;
    lbct_link *link;
    struct sockaddr *dest;
    LBCTailgatePacketGroup *pg;
    LBCTailgatePacketGroupDef pg_def;
    KLListNode *list_node;

    /* Trim the tailgate list. */
    if (KLListCount(lbc_tailgate->packet_group_list) > 100) {
	for (i = 0; i < 50; i++) {
	    list_node = KLListFirst(lbc_tailgate->packet_group_list);
	    pg = KLListGetItem(list_node);
	    KLListFreeNode(lbc_tailgate->packet_group_list, list_node);
	    lbc_tailgate_packet_group_free(pg);
	}
    }

    /* Find link to measure. */
    tailgated_ttl = lbct_select_link(lbc_tailgate);

    /* Try to find size with smallest count. */
    link = lbct_link_search(lbc_tailgate->link_list, tailgated_ttl);
    if (link != NULL) {
	lbc_packet_size_range_find_min_count(
	    link->ps_range, &packet_count, &tailgater_size);
    }
    else {
	tailgater_size = lbc_tailgate->min_packet_size;
    }
    /* printf("Tailgater size: %d\n", tailgater_size); */

    /* Set up packet group */
    pg_def.num_packets = 2;
    pg_def.num_tailgated_packets = 1;
    pg_def.packet_id = lbc_tailgate->packet_id;
    pg_def.ttl = tailgated_ttl;
    pg_def.tailgater_size = tailgater_size;
    pg = lbc_tailgate_packet_group_new(&pg_def);
    KLListAppendItem(lbc_tailgate->packet_group_list, pg);

    /* Prepare the tailgated for sending. */
    dest = &lbc_tailgate->def.flow.dest.sock_addr;
#if 0
    /* XXX Don't rely on ttl expiration. */
    if ((link != NULL) && (link->addr_set) && 0) {
	dest = &link->addr;
	protocol = IPPROTO_UDP;
    }
#endif
    size = lbc_tailgate->max_packet_size;
    transport_compose_packet(
	lbc_tailgate->transport_session, dest,
	lbc_tailgate->def.flow.ip_protocol,
	tailgated_ttl, size, lbc_tailgate->packet_id, lbc_tailgate->buf0);
    lbc_tailgate->packet_id += 1;

    /* Prepare the tailgater for sending. */
    size = tailgater_size;
    transport_compose_packet(
	lbc_tailgate->transport_session, dest,
	lbc_tailgate->def.flow.ip_protocol,
	255, size, lbc_tailgate->packet_id, lbc_tailgate->buf1);
    lbc_tailgate->packet_id += 1;

    /* Send them. */
    lbc_tailgate->max_packet_size = transport_send_packet(
	lbc_tailgate->transport_session, lbc_tailgate->buf0);
    lbc_tailgate->min_packet_size = transport_send_packet(
	lbc_tailgate->transport_session, lbc_tailgate->buf1);

    lbc_tailgate->packets_sent += 2;
    return 2;
}

void
lbc_links_get_string(
    LBCTailgate *lbc_tailgate, double sigma_bandwidth, double sigma_delay,
    KLArray *array)
{
    char src_string[INET_ADDRSTRLEN], dst_string[INET_ADDRSTRLEN];
    lbct_link *link = NULL;
    KLListNode *list_node;
    double partial_sigma = 0.0, last_inv_bw, min_delay = 0.0;

    KLListLoop(list_node, lbc_tailgate->link_list) {
	link = KLListGetItem(list_node);

	/* Get endpoints */
	lbct_link_get_endpoints(
	    lbc_tailgate, list_node, src_string, sizeof(src_string),
	    dst_string, sizeof(dst_string));
	KLArrayStrCat(array, "%-16s -> %-16s", src_string, dst_string);

	/* Calculate bandwidth. */
	lbct_link_calc(lbc_tailgate, link, sigma_bandwidth, sigma_delay);
	min_delay = KLTimingGetDouble(&link->min_delay);
	KLArrayStrCat(
	    array, "  ttl:%3d  bw:%12.0f  packets:%5d\n",
	    link->ttl, (1 / link->inv_bw), link->packets_received);
	KLArrayStrCat(
	    array, "  rtt:%9.9f  error:%6.6f  mean:%9.9f\n",
	    min_delay, link->min_delay_error, link->min_delay_mean);
	KLArrayStrCat(
	    array, "  ack_size:%3d  md0: %5d  md1:%5d  ack_md:%5d\n",
	    link->ack_size, link->md0, link->md1, link->md_ack);
	lbc_packet_size_range_get_string(link->ps_range, array);
	partial_sigma += link->inv_bw;
    }
    if (link != NULL) {
	partial_sigma -= link->inv_bw;
	last_inv_bw = sigma_bandwidth - partial_sigma;
	KLArrayStrCat(array, "%-16s -> %-16s bw: %9.0f rtt: %.6f\n",
		    src_string, dst_string, (1 / last_inv_bw), min_delay);
    }
}

void
lbc_tailgate_get_max_error(LBCTailgate *lbc_tailgate, double *max_error_p)
{
    KLListNode *list_node;
    lbct_link *link;
    double max_error = 0.0;
    
    KLListLoop(list_node, lbc_tailgate->link_list) {
	link = KLListGetItem(list_node);
	lbct_link_bootstrap(lbc_tailgate, link);
	if (link->min_delay_error > max_error) {
	    max_error = link->min_delay_error;
	}
    }
    *max_error_p = max_error;
}

/*
 * Handle the arrival of a new sample.
 */
int
lbc_tailgate_sample_received(
    LBCTailgate *lbc_tailgate, const DPCapFlowPacket *packet, 
    const struct timespec *delay, unsigned int timing_type)
{
    LBCTailgatePacketGroup *pg;
    KLListNode *pg_list_node;
    lbct_link *link;
    unsigned int packet_ttl, packet_id, min_packet_size, max_packet_size;
    const KLNetworkProtocolPacket *np_p;

    np_p = DPCapFlowPacketGetPacket(packet);
    packet_id = transport_get_packet_id(
	lbc_tailgate->transport_session,
	&lbc_tailgate->def.flow.dest.sock_addr, np_p);
    pg_list_node = lbct_check_packet(
	lbc_tailgate, packet, packet_id, timing_type);
    if (pg_list_node == NULL) {
	return 0;
    }
    pg = KLListGetItem(pg_list_node);
    lbc_tailgate_packet_group_new_sample(pg, packet, packet_id, delay);
    /* printf("Packet id: %d\n", packet_id); */
    if (timing_type == 0) {
	return 0;
    }
    lbc_tailgate->packets_received++;
    packet_ttl = lbc_tailgate_packet_group_get_ttl(pg);
    link = lbct_link_search(lbc_tailgate->link_list, packet_ttl);
    if ((link == NULL) && !lbc_tailgate->dest_found &&
	(KLListCount(lbc_tailgate->link_list) < 64)) {
	/*
	 * Need to create a new link and the destination hasn't been
	 * found and still have space for more links.
	 */
	transport_session_get_packet_sizes(
	    lbc_tailgate->transport_session, IPPROTO_TCP, &min_packet_size,
	    &max_packet_size);
	link = lbct_link_new(packet_ttl, min_packet_size, max_packet_size);
	KLListAppendItem(lbc_tailgate->link_list, link);
    }
    link->packets_received++;

    lbct_link_set_addr(lbc_tailgate, link, pg);

    /* Check if we have timings for all packets in this group. */
    if (lbc_tailgate_packet_group_check_finished(pg)) {
	lbct_link_add_sample(link, pg);
	KLListFreeNode(lbc_tailgate->packet_group_list, pg_list_node);
	lbc_tailgate_packet_group_free(pg);
	return 1;
    }
    return 0;
}

